import asyncio

import asynctest
from unittest import mock
from uuid import uuid4

from pulpcore.plugin.stages import DeclarativeContent, DeclarativeArtifact
from pulpcore.plugin.stages.artifact_stages import ArtifactDownloader


class MockException(Exception):
    """
    A tracer exception.
    """

    pass


class DownloaderMock:
    """Mock for a Downloader.

    URLs are expected to be the delay to wait to simulate downloading,
    e.g `url='5'` will wait for 5 seconds. Negative numbers will raise
    an exception after waiting for the absolute value, e.g. `url=-5` fails
    after 5 seconds. `DownloaderMock` manages _global_ statistics about the
    downloads.
    """

    running = 0
    downloads = 0
    canceled = 0

    def __init__(self, url, **kwargs):
        self.url = url

    @classmethod
    def reset(cls):
        cls.running = 0
        cls.downloads = 0
        cls.canceled = 0

    async def run(self, extra_data=None):
        DownloaderMock.running += 1
        try:
            duration = int(self.url)
            await asyncio.sleep(abs(duration))
            if duration < 0:
                raise MockException("Download Failed")
        except asyncio.CancelledError:
            DownloaderMock.running -= 1
            DownloaderMock.canceled += 1
            raise
        DownloaderMock.running -= 1
        DownloaderMock.downloads += 1
        result = mock.Mock()
        result.url = self.url
        result.artifact_attributes = {}
        return result


class TestArtifactDownloader(asynctest.ClockedTestCase):
    def setUp(self):
        super().setUp()
        DownloaderMock.reset()
        self.now = 0
        self.in_q = asyncio.Queue()
        self.out_q = asyncio.Queue()

    async def advance_to(self, now):
        delta = now - self.now
        assert delta >= 0
        await self.advance(delta)
        self.now = now

    async def advance(self, delta):
        await super().advance(delta)
        self.now += delta

    def queue_dc(self, delays=[], artifact_path=None):
        """Put a DeclarativeContent instance into `in_q`

        For each `delay` in `delays`, associate a DeclarativeArtifact
        with download duration `delay` to the content unit. `delay ==
        None` means that the artifact is already present (pk is set)
        and no download is required. `artifact_path != None` means
        that the Artifact already has a file associated with it and a
        download does not need to be scheduled.
        """
        das = []
        for delay in delays:
            artifact = mock.Mock()
            artifact.pk = uuid4()
            artifact._state.adding = delay is not None
            artifact.DIGEST_FIELDS = []
            artifact.file = artifact_path
            remote = mock.Mock()
            remote.get_downloader = DownloaderMock
            das.append(
                DeclarativeArtifact(
                    artifact=artifact, url=str(delay), relative_path="path", remote=remote
                )
            )
        dc = DeclarativeContent(content=mock.Mock(), d_artifacts=das)
        self.in_q.put_nowait(dc)

    async def download_task(self, max_concurrent_content=3):
        """
        A coroutine running the downloader stage with a mocked ProgressReport.

        Returns:
            The done count of the ProgressReport.
        """
        with mock.patch("pulpcore.plugin.stages.artifact_stages.ProgressReport") as pb:
            pb.return_value.__aenter__.return_value.done = 0
            ad = ArtifactDownloader(max_concurrent_content=max_concurrent_content)
            ad._connect(self.in_q, self.out_q)
            await ad()
        return pb.return_value.__aenter__.return_value.done

    def assertQueued(self, num):
        self.assertEqual(self.in_q.qsize(), num)

    def assertHandled(self, num):
        self.assertEqual(self.out_q.qsize(), num)

    async def test_downloads(self):
        download_task = self.loop.create_task(self.download_task())

        # Create 28 content units, every third one must be downloaded.
        # The downloads take 0, 3, 6,..., 27 seconds; content units
        # 1, 2, 4, 5, ..., 26 do not need downloads.
        for i in range(28):
            self.queue_dc(delays=[i if not i % 3 else None])
        self.in_q.put_nowait(None)

        # At 0.5 seconds
        await self.advance_to(0.5)
        # 3, 6 and 9 are running. 0 is finished
        self.assertEqual(DownloaderMock.running, 3)
        # non-downloads 1, 2, 4, 5, 7, 8 are forwarded
        self.assertHandled(7)
        # 9 - 26 + None are waiting to be picked up
        self.assertQueued(19)

        # Two downloads run in parallel. The most asymmetric way
        # to schedule the remaining downloads is:
        # 3 + 12 + 21: finished after 36 seconds
        # 6 + 15 + 24: finished after 45 seconds
        # 9 + 18 + 27: finished after 54 seconds
        for t in range(1, 36):  # until 35.5 seconds three downloads must run
            await self.advance_to(t + 0.5)
            self.assertEqual(DownloaderMock.running, 3)

        # At 54.5 seconds, the stage is done at the latest
        await self.advance_to(54.5)
        self.assertEqual(DownloaderMock.running, 0)
        self.assertEqual(DownloaderMock.downloads, 10)
        self.assertEqual(download_task.result(), DownloaderMock.downloads)
        self.assertQueued(0)
        self.assertHandled(29)

    async def test_multi_artifact_downloads(self):
        # Content units should fill the slot like
        #
        # 0   1   2   3 s
        # .   .   .   .
        # +---+-------+
        # | 1 |   4   |
        # +---+---+---+
        # |   2   |
        # +-------+
        # |   3   |
        # +-------+
        #
        download_task = self.loop.create_task(self.download_task())
        self.queue_dc(delays=[])  # must be forwarded to next stage immediately
        self.queue_dc(delays=[1])
        self.queue_dc(delays=[2, 2])
        self.queue_dc(delays=[2])
        self.queue_dc(delays=[2, None])  # schedules only one download
        self.in_q.put_nowait(None)
        # At 0.5 seconds, three content units are downloading with four
        # downloads overall
        await self.advance_to(0.5)
        self.assertEqual(DownloaderMock.running, 4)
        self.assertHandled(1)
        # At 1.5 seconds, the download for the first content unit has completed.
        # At 1 second, the download of the forth content unit is started
        await self.advance_to(1.5)
        self.assertEqual(DownloaderMock.running, 4)
        self.assertHandled(2)
        # At 2.5 seconds, the downloads for the second and the third content unit
        # have completed
        await self.advance_to(2.5)
        self.assertEqual(DownloaderMock.running, 1)
        self.assertHandled(4)

        # At 3.5 seconds, stage must de done
        await self.advance_to(3.5)
        self.assertEqual(DownloaderMock.running, 0)
        self.assertEqual(DownloaderMock.downloads, 5)
        self.assertEqual(download_task.result(), DownloaderMock.downloads)
        self.assertQueued(0)
        self.assertHandled(6)

    async def test_sparse_batches_dont_block_stage(self):
        """Regression test for issue https://pulp.plan.io/issues/4018."""

        def queue_content_with_a_single_download(batchsize=100, delay=100):
            """
            Queue a batch of `batchsize` declarative_content instances. Only the
            first one triggers a download of duration `delay`.
            """
            self.queue_dc(delays=[delay])
            for i in range(batchsize - 1):
                self.queue_dc([None])

        download_task = self.loop.create_task(self.download_task())

        queue_content_with_a_single_download()

        # At 0.5 seconds, the first content unit is downloading
        await self.advance_to(0.5)
        self.assertEqual(DownloaderMock.running, 1)
        self.assertHandled(99)

        # at 0.5 seconds next batch arrives (last batch)
        queue_content_with_a_single_download()
        self.in_q.put_nowait(None)

        # at 1.0 seconds, two downloads are running
        await self.advance_to(1)
        self.assertEqual(DownloaderMock.running, 2)
        self.assertHandled(2 * 99)

        # at 101 seconds, stage should have completed
        await self.advance_to(101)

        self.assertEqual(DownloaderMock.running, 0)
        self.assertEqual(DownloaderMock.downloads, 2)
        self.assertEqual(download_task.result(), DownloaderMock.downloads)
        self.assertQueued(0)
        self.assertHandled(201)

    async def test_cancel(self):
        download_task = self.loop.create_task(self.download_task())
        for i in range(4):
            self.queue_dc(delays=[100])
        self.in_q.put_nowait(None)

        # After 0.5 seconds, the three downloads must have started
        await self.advance_to(0.5)
        self.assertEqual(DownloaderMock.running, 3)

        download_task.cancel()

        await self.advance_to(1.0)

        with self.assertRaises(asyncio.CancelledError):
            download_task.result()
        self.assertEqual(DownloaderMock.running, 0)
        self.assertEqual(DownloaderMock.canceled, 3)

    async def test_exception_with_empty_in_q(self):
        download_task = self.loop.create_task(self.download_task())

        # Create three content units with 1 downloads, followed by one thowing an exception.
        self.queue_dc(delays=[1])
        self.queue_dc(delays=[2])
        self.queue_dc(delays=[2])
        self.queue_dc(delays=[-1])

        # At 0.5 seconds
        await self.advance_to(0.5)
        # 3 downloads are running. No unit is finished
        self.assertEqual(DownloaderMock.running, 3)
        self.assertHandled(0)
        self.assertQueued(1)

        # At 1.5 seconds
        await self.advance_to(1.5)
        # 3 downloads are running. One unit is finished.
        self.assertEqual(DownloaderMock.running, 3)
        self.assertHandled(1)
        self.assertQueued(0)

        # At 2.5 seconds, the exception must have been triggered
        await self.advance_to(2.5)
        self.assertTrue(download_task.done())
        self.assertIsInstance(download_task.exception(), MockException)

    async def test_exception_finished_in_q(self):
        download_task = self.loop.create_task(self.download_task())

        # Create three content units with 1 downloads, followed by one thowing an exception.
        self.queue_dc(delays=[1])
        self.queue_dc(delays=[2])
        self.queue_dc(delays=[2])
        self.queue_dc(delays=[-1])
        await self.in_q.put(None)

        # At 0.5 seconds
        await self.advance_to(0.5)
        # 3 downloads are running. No unit is finished
        self.assertEqual(DownloaderMock.running, 3)
        self.assertHandled(0)
        self.assertQueued(2)

        # At 1.5 seconds
        await self.advance_to(1.5)
        # 3 downloads are running. One unit is finished.
        self.assertEqual(DownloaderMock.running, 3)
        self.assertHandled(1)
        self.assertQueued(1)

        # At 2.5 seconds, the exception must have been triggered
        await self.advance_to(2.5)
        self.assertTrue(download_task.done())
        self.assertIsInstance(download_task.exception(), MockException)

    async def test_exception_with_saturated_content_slots(self):
        download_task = self.loop.create_task(self.download_task())

        # Create three content units with 1 downloads, followed by one thowing an exception.
        self.queue_dc(delays=[1])
        self.queue_dc(delays=[3])
        self.queue_dc(delays=[3])
        self.queue_dc(delays=[-1])
        self.queue_dc(delays=[1])  # This unit will be waiting for a free slot

        # At 0.5 seconds
        await self.advance_to(0.5)
        # 3 downloads are running. No unit is finished
        self.assertEqual(DownloaderMock.running, 3)
        self.assertHandled(0)
        self.assertQueued(2)

        # At 1.5 seconds
        await self.advance_to(1.5)
        # 3 downloads are running. One unit is finished.
        self.assertEqual(DownloaderMock.running, 3)
        self.assertHandled(1)
        self.assertQueued(1)

        # At 2.5 seconds, the exception must have been triggered
        await self.advance_to(2.5)
        self.assertTrue(download_task.done())
        self.assertIsInstance(download_task.exception(), MockException)

    async def test_download_artifact_with_file(self):
        download_task = self.loop.create_task(self.download_task())

        # Create 3 downloads with Artifacts that already have a file
        self.queue_dc(delays=[1, 2, 3], artifact_path="/foo/bar")
        # Create 3 downloads with Artifacts that don't have a file
        self.queue_dc(delays=[1, 2, 3], artifact_path=None)
        self.in_q.put_nowait(None)

        # At 0.5 seconds only 3 should be running and 0 done
        await self.advance_to(0.5)
        self.assertEqual(DownloaderMock.downloads, 0)
        self.assertEqual(DownloaderMock.running, 3)

        # At 10 seconds all 3 should be done
        await self.advance_to(3)
        self.assertEqual(DownloaderMock.downloads, 3)
        self.assertEqual(DownloaderMock.running, 0)
        self.assertEqual(download_task.result(), DownloaderMock.downloads)



import gc

from django.conf import settings

from pulpcore.app.models import (
    Artifact,
    Content,
    ProgressReport,
    PublishedMetadata,
)


def queryset_iterator(qs, batchsize=2000, gc_collect=True):
    """
    Provide a batching functionality that returns individual querysets per batch.

    Copied from here with minor changes:
    https://www.guguweb.com/2020/03/27/optimize-django-memory-usage/
    """
    iterator = qs.values_list("pk", flat=True).order_by("pk").distinct().iterator()
    eof = False
    while not eof:
        primary_key_buffer = []
        try:
            while len(primary_key_buffer) < batchsize:
                primary_key_buffer.append(next(iterator))
        except StopIteration:
            eof = True
        yield qs.filter(pk__in=primary_key_buffer).order_by("pk")
        if gc_collect:
            gc.collect()


def orphan_cleanup(content_pks=None, orphan_protection_time=settings.ORPHAN_PROTECTION_TIME):
    """
    Delete all orphan Content and Artifact records.
    Go through orphan Content multiple times to remove content from subrepos.
    This task removes Artifact files from the filesystem as well.

    Kwargs:
        content_pks (list): A list of content pks. If specified, only remove these orphans.

    """
    progress_bar = ProgressReport(
        message="Clean up orphan Content",
        total=0,
        code="clean-up.content",
        done=0,
        state="running",
    )

    while True:
        content = Content.objects.orphaned(orphan_protection_time, content_pks).exclude(
            pulp_type=PublishedMetadata.get_pulp_type()
        )
        content_count = content.count()
        if not content_count:
            break

        progress_bar.total += content_count
        progress_bar.save()

        # delete the content
        for c in queryset_iterator(content):
            progress_bar.increase_by(c.count())
            c.delete()

    progress_bar.state = "completed"
    progress_bar.save()

    # delete the artifacts that don't belong to any content
    artifacts = Artifact.objects.orphaned(orphan_protection_time)

    progress_bar = ProgressReport(
        message="Clean up orphan Artifacts",
        total=artifacts.count(),
        code="clean-up.content",
        done=0,
        state="running",
    )
    progress_bar.save()

    counter = 0
    interval = 100
    for artifact in artifacts.iterator():
        # we need to manually call delete() because it cleans up the file on the filesystem
        artifact.delete()
        progress_bar.done += 1
        counter += 1

        if counter >= interval:
            progress_bar.save()
            counter = 0

    progress_bar.state = "completed"
    progress_bar.save()



# Generated by Django 2.2.13 on 2020-08-07 18:51

from django.db import migrations, models, transaction


def update_download_concurrency(apps, schema_editor):
    with transaction.atomic():
        Remote = apps.get_model('core', 'Remote')
        Remote.objects.filter(download_concurrency=20).update(download_concurrency=10)


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0038_repository_remote'),
    ]

    operations = [
        migrations.AlterField(
            model_name='remote',
            name='download_concurrency',
            field=models.PositiveIntegerField(default=10),
        ),
        migrations.RunPython(update_download_concurrency),
    ]



import asyncio
from unittest import skip
from unittest.mock import patch, call

from django.db.utils import InterfaceError, OperationalError
from django.test import TestCase

from pulpcore.content import _heartbeat


class ContentHeartbeatTestCase(TestCase):
    @skip("Skipping while resolving https://github.com/rochacbruno/dynaconf/issues/689")
    @patch("pulpcore.app.models.ContentAppStatus.objects.get_or_create")
    @patch("pulpcore.content.handler.Handler._reset_db_connection")
    def test_db_connection_interface_error(self, mock_reset_db, mock_get_or_create):
        """
        Test that if an InterfaceError or OperationalError is raised,
        Handler._reset_db_connection() is called
        """

        class MockException(Exception):
            pass

        mock_get_or_create.side_effect = [InterfaceError(), OperationalError(), MockException()]

        loop = asyncio.get_event_loop()
        with self.settings(CONTENT_APP_TTL=1):
            try:
                loop.run_until_complete(_heartbeat())
            except MockException:
                pass
        loop.close()

        mock_get_or_create.assert_called()
        mock_reset_db.assert_has_calls([call(), call()])



from gettext import gettext as _

from rest_framework import fields, serializers

from pulpcore.app.serializers import ValidateFieldsMixin  # noqa


class RepairSerializer(serializers.Serializer, ValidateFieldsMixin):
    verify_checksums = fields.BooleanField(
        required=False,
        default=True,
        help_text=_(
            "Will verify that the checksum of all stored files matches what saved in the "
            "database. Otherwise only the existence of the files will be checked. Enabled "
            "by default"
        ),
    )



import atexit
import sys
import coverage

cov = coverage.Coverage()
cov.start()


def finalize_coverage():
    print("writing coverage", file=sys.stderr)
    cov.stop()
    cov.save()


atexit.register(finalize_coverage)

from ...content import server  # noqa



from .api import create_pipeline, EndStage, Stage  # noqa
from .artifact_stages import (  # noqa
    ACSArtifactHandler,
    ArtifactDownloader,
    ArtifactSaver,
    QueryExistingArtifacts,
    RemoteArtifactSaver,
)
from .content_stages import (  # noqa
    ContentAssociation,
    ContentSaver,
    QueryExistingContents,
    ResolveContentFutures,
)
from .declarative_version import DeclarativeVersion  # noqa
from .models import DeclarativeArtifact, DeclarativeContent  # noqa
from .profiler import ProfilingQueue, create_profile_db_and_connection  # noqa



# Generated by Django 2.2.18 on 2021-02-02 21:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0057_add_label_indexes'),
    ]

    operations = [
        migrations.AddField(
            model_name='accesspolicy',
            name='customized',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.24 on 2021-07-08 12:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0069_update_json_fields'),
    ]

    operations = [
        migrations.RenameField(
            model_name='repository',
            old_name='retained_versions',
            new_name='retain_repo_versions',
        ),
    ]



import unittest

from pulp_smash import config
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task

from pulpcore.client.pulp_file import AcsFileApi, RemotesFileApi
from pulpcore.client.pulp_file.exceptions import ApiException

from pulpcore.tests.functional.api.using_plugin.utils import (
    gen_file_client,
    gen_file_remote,
)


class AlternateContentSourceTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        Create class-wide variables.

        Variables 'paths' and 'paths_updated' are defined as strings.
        In same way data are send from user.
        """
        cls.cfg = config.get_config()
        cls.file_client = gen_file_client()
        cls.file_remote_api = RemotesFileApi(cls.file_client)
        cls.file_acs_api = AcsFileApi(cls.file_client)
        cls.paths = ["backupone/PULP_MANIFEST", "backuptwo/manifest"]
        cls.paths_updated = ["backupone/test", "anotherbackup/PULP_MANIFEST"]

    @classmethod
    def tearDownClass(cls):
        delete_orphans()

    def test_create(self):
        """
        Basic ACS create.

        1. Try and fail to create ACS with remote with immediate policy
        2. Create ACS and check it exists
        """
        remote_bad = self.file_remote_api.create(gen_file_remote())
        remote = self.file_remote_api.create(gen_file_remote(policy="on_demand"))
        self.addCleanup(self.file_remote_api.delete, remote_bad.pulp_href)
        self.addCleanup(self.file_remote_api.delete, remote.pulp_href)

        acs_data = {
            "name": "alternatecontentsource",
            "remote": remote_bad.pulp_href,
            "paths": self.paths,
        }
        with self.assertRaises(ApiException) as ctx:
            self.file_acs_api.create(acs_data)
        self.assertEqual(ctx.exception.status, 400)

        acs_data["remote"] = remote.pulp_href

        acs = self.file_acs_api.create(acs_data)
        self.addCleanup(self.file_acs_api.delete, acs.pulp_href)

        self.assertEqual(len(self.file_acs_api.list(name="alternatecontentsource").results), 1)

    def test_acs_update(self):
        """
        ACS update.

        Test of update name and paths.
        """
        remote = self.file_remote_api.create(gen_file_remote(policy="on_demand"))
        self.addCleanup(self.file_remote_api.delete, remote.pulp_href)

        acs_data = {
            "name": "alternatecontentsource",
            "remote": remote.pulp_href,
            "paths": self.paths,
        }
        acs = self.file_acs_api.create(acs_data)
        self.addCleanup(self.file_acs_api.delete, acs.pulp_href)

        # update name
        new_name = "acs"
        response = self.file_acs_api.update(acs.pulp_href, {"name": new_name, "remote": acs.remote})
        monitor_task(response.task)
        acs = self.file_acs_api.read(acs.pulp_href)

        self.assertEqual(acs.name, new_name)
        # assert paths were not silently removed during name update
        self.assertEqual(sorted(acs.paths), sorted(self.paths))

        # partial update name
        new_name = "new_acs"
        response = self.file_acs_api.partial_update(
            acs.pulp_href, {"name": new_name, "remote": acs.remote}
        )
        monitor_task(response.task)
        acs = self.file_acs_api.read(acs.pulp_href)

        self.assertEqual(acs.name, new_name)

        # update paths
        response = self.file_acs_api.update(
            acs.pulp_href, {"name": acs.name, "remote": acs.remote, "paths": self.paths_updated}
        )
        monitor_task(response.task)
        acs = self.file_acs_api.read(acs.pulp_href)

        self.assertEqual(sorted(acs.paths), sorted(self.paths_updated))



import re

from django import template
from django.utils.encoding import force_text
from django.utils.html import escape, smart_urlquote
from django.utils.safestring import SafeData, mark_safe

register = template.Library()


TRAILING_PUNCTUATION = [".", ",", ":", ".)", '"', "'", "&quot;", ";"]
WRAPPING_PUNCTUATION = [
    ("(", ")"),
    ("<", ">"),
    ("[", "]"),
    ("&lt;", "&gt;"),
    ('"', '"'),
    ("'", "'"),
    ("&quot;", "&quot;"),
]
word_split_re = re.compile(r"(\s+)")
href_re = re.compile(r"^\/pulp\/api\/v3\/", re.IGNORECASE)


@register.filter(needs_autoescape=True)
def urlize_quoted_hrefs(text, trim_url_limit=None, nofollow=True, autoescape=True):
    """
    Converts href in text into clickable links. If trim_url_limit is not None, the URLs in link
    text longer than this limit will truncated to trim_url_limit-3 characters and appended with
    an ellipsis. If nofollow is True, the URLs in link text will get a rel="nofollow" attribute.

    """

    def trim_url(x, limit=trim_url_limit):
        return limit is not None and (len(x) > limit and ("%s..." % x[: max(0, limit - 3)])) or x

    safe_input = isinstance(text, SafeData)
    words = word_split_re.split(force_text(text))
    for i, word in enumerate(words):
        if "/pulp/api/v3/" in word:
            # Deal with punctuation.
            lead, middle, trail = "", word, ""
            for punctuation in TRAILING_PUNCTUATION:
                if middle.endswith(punctuation):
                    middle = middle[: -len(punctuation)]
                    trail = punctuation + trail
            for opening, closing in WRAPPING_PUNCTUATION:
                if middle.startswith(opening):
                    middle = middle[len(opening) :]
                    lead = lead + opening
                # Keep parentheses at the end only if they're balanced.
                if middle.endswith(closing) and middle.count(closing) == middle.count(opening) + 1:
                    middle = middle[: -len(closing)]
                    trail = closing + trail

            # Make URL we want to point to.
            url = None
            nofollow_attr = ' rel="nofollow"' if nofollow else ""

            if href_re.match(middle):
                url = smart_urlquote(middle)

            # Check if it's a real URL
            if url and ("{" in url or "%7B" in url):
                url = None

            # Make link.
            if url:
                trimmed = trim_url(middle)
                if autoescape and not safe_input:
                    lead, trail = escape(lead), escape(trail)
                    url, trimmed = escape(url), escape(trimmed)
                middle = '<a href="%s"%s>%s</a>' % (url, nofollow_attr, trimmed)
                words[i] = mark_safe("%s%s%s" % (lead, middle, trail))
            else:
                if safe_input:
                    words[i] = mark_safe(word)
                elif autoescape:
                    words[i] = escape(word)
        elif safe_input:
            words[i] = mark_safe(word)
        elif autoescape:
            words[i] = escape(word)
    return "".join(words)



"""
Container for models using generic relations provided by Django's ContentTypes framework.

References:
    https://docs.djangoproject.com/en/1.8/ref/contrib/contenttypes/#generic-relations
"""

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

from pulpcore.app.models.base import BaseModel


class GenericRelationModel(BaseModel):
    """Base model class for implementing Generic Relations.

    This class provides the required fields to implement generic relations. Instances of
    this class can only be related models with a primary key, such as those subclassing
    Pulp's base Model class.
    """

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.UUIDField()
    content_object = GenericForeignKey("content_type", "object_id")

    class Meta:
        abstract = True



from .base import (  # noqa
    AsyncCreateMixin,
    AsyncRemoveMixin,
    AsyncUpdateMixin,
    BaseFilterSet,
    NamedModelViewSet,
    NAME_FILTER_OPTIONS,
)

from .access_policy import AccessPolicyViewSet  # noqa

from .acs import AlternateContentSourceViewSet  # noqa

from .content import (  # noqa
    ArtifactFilter,
    ArtifactViewSet,
    ContentFilter,
    ContentViewSet,
    ListContentViewSet,
    ReadOnlyContentViewSet,
    SigningServiceViewSet,
)
from .custom_filters import (  # noqa
    IsoDateTimeFilter,
    RepoVersionHrefFilter,
    RepositoryVersionFilter,
)
from .exporter import (  # noqa
    ExportViewSet,
    ExporterViewSet,
    FilesystemExporterViewSet,
    FilesystemExportViewSet,
    PulpExporterViewSet,
    PulpExportViewSet,
)
from .importer import (  # noqa
    ImportViewSet,
    ImporterViewSet,
    PulpImportViewSet,
    PulpImporterViewSet,
)
from .orphans import OrphansCleanupViewset  # noqa
from .publication import (  # noqa
    ContentGuardFilter,
    ContentGuardViewSet,
    DistributionFilter,
    DistributionViewSet,
    ListContentGuardViewSet,
    ListPublicationViewSet,
    PublicationFilter,
    PublicationViewSet,
    RBACContentGuardViewSet,
)
from .reclaim import ReclaimSpaceViewSet  # noqa
from .repository import (  # noqa
    ImmutableRepositoryViewSet,
    ListRepositoryViewSet,
    ReadOnlyRepositoryViewSet,
    RemoteFilter,
    RemoteViewSet,
    RepositoryViewSet,
    RepositoryVersionViewSet,
    ListRepositoryVersionViewSet,
)
from .task import TaskViewSet, TaskGroupViewSet, WorkerViewSet  # noqa
from .upload import UploadViewSet  # noqa
from .user import (  # noqa
    GroupViewSet,
    GroupUserViewSet,
    GroupModelPermissionViewSet,
    GroupObjectPermissionViewSet,
    UserViewSet,
)



import os
from gettext import gettext as _

from rest_framework import serializers
from rest_framework.validators import UniqueValidator

from pulpcore.app import models, settings
from pulpcore.app.serializers import (
    DetailIdentityField,
    ImportIdentityField,
    ModelSerializer,
    RelatedField,
    ValidateFieldsMixin,
)


class ImporterSerializer(ModelSerializer):
    """Base serializer for Importers."""

    pulp_href = DetailIdentityField(view_name_pattern=r"importer(-.*/.*)-detail")
    name = serializers.CharField(
        help_text=_("Unique name of the Importer."),
        validators=[UniqueValidator(queryset=models.Importer.objects.all())],
    )

    class Meta:
        model = models.Importer
        fields = ModelSerializer.Meta.fields + ("name",)


class ImportSerializer(ModelSerializer):
    """Serializer for Imports."""

    pulp_href = ImportIdentityField()

    task = RelatedField(
        help_text=_("A URI of the Task that ran the Import."),
        queryset=models.Task.objects.all(),
        view_name="tasks-detail",
    )

    params = serializers.JSONField(
        help_text=_("Any parameters that were used to create the import."),
    )

    class Meta:
        model = models.Importer
        fields = ModelSerializer.Meta.fields + ("task", "params")


class PulpImporterSerializer(ImporterSerializer):
    """Serializer for PulpImporters."""

    repo_mapping = serializers.DictField(
        child=serializers.CharField(),
        help_text=_(
            "Mapping of repo names in an export file to the repo names in Pulp. "
            "For example, if the export has a repo named 'foo' and the repo to "
            "import content into was 'bar', the mapping would be \"{'foo': 'bar'}\"."
        ),
        required=False,
    )

    def create(self, validated_data):
        """
        Save the PulpImporter and handle saving repo mapping.

        Args:
            validated_data (dict): A dict of validated data to create the PulpImporter

        Raises:
            ValidationError: When there's a problem with the repo mapping.

        Returns:
            PulpImporter: the created PulpImporter
        """
        repo_mapping = validated_data.pop("repo_mapping", {})
        importer = super().create(validated_data)
        try:
            importer.repo_mapping = repo_mapping
        except Exception as err:
            importer.delete()
            raise serializers.ValidationError(_("Bad repo mapping: {}").format(err))
        else:
            return importer

    class Meta:
        model = models.PulpImporter
        fields = ImporterSerializer.Meta.fields + ("repo_mapping",)


class PulpImportSerializer(ModelSerializer):
    """Serializer for call to import into Pulp."""

    path = serializers.CharField(
        help_text=_("Path to export that will be imported."), required=False
    )
    toc = serializers.CharField(
        help_text=_(
            "Path to a table-of-contents file describing chunks to be validated, "
            + "reassembled, and imported."
        ),
        required=False,
    )

    def _check_path_allowed(self, param, a_path):
        user_provided_realpath = os.path.realpath(a_path)
        for allowed_path in settings.ALLOWED_IMPORT_PATHS:
            if user_provided_realpath.startswith(allowed_path):
                return user_provided_realpath

        raise serializers.ValidationError(
            _("{} '{}' is not an allowed import path").format(param, a_path)
        )

    def validate_path(self, value):
        """
        Check if path exists and is in ALLOWED_IMPORT_PATHS.

        Args:
            value (str): The user-provided value path to be validated.

        Raises:
            ValidationError: When path is not in the ALLOWED_IMPORT_PATHS setting.

        Returns:
            The validated value.
        """
        return self._check_path_allowed("path", value)

    def validate_toc(self, value):
        """
        Check validity of provided 'toc' parameter.

        'toc' must be within ALLOWED_IMPORT_PATHS.

        NOTE: this method does NOT validate existence/sanity of export-files. That
        happens asynchronously, due to time/responsiveness constraints.

        Args:
            value (str): The user-provided toc-file-path to be validated.

        Raises:
            ValidationError: When toc is not in the ALLOWED_IMPORT_PATHS setting

        Returns:
            The validated value.
        """

        return self._check_path_allowed("toc", value)

    def validate(self, data):
        # only one-of 'path'/'toc'
        if data.get("path", None) and data.get("toc", None):
            raise serializers.ValidationError(_("Only one of 'path' and 'toc' may be specified."))

        # requires one-of 'path'/'toc'
        if not data.get("path", None) and not data.get("toc", None):
            raise serializers.ValidationError(_("One of 'path' or 'toc' must be specified."))

        return super().validate(data)

    class Meta:
        model = models.Import
        fields = (
            "path",
            "toc",
        )


class EvaluationSerializer(serializers.Serializer):
    """
    Results from evaluating a proposed parameter to a PulpImport call.
    """

    context = serializers.CharField(
        help_text=_("Parameter value being evaluated."),
    )
    is_valid = serializers.BooleanField(
        help_text=_("True if evaluation passed, false otherwise."),
    )
    messages = serializers.ListField(
        child=serializers.CharField(),
        help_text=_("Messages describing results of all evaluations done. May be an empty list."),
    )


class PulpImportCheckResponseSerializer(serializers.Serializer):
    """
    Return the response to a PulpImport import-check call.
    """

    toc = EvaluationSerializer(
        help_text=_("Evaluation of proposed 'toc' file for PulpImport"),
        required=False,
    )
    path = EvaluationSerializer(
        help_text=_("Evaluation of proposed 'path' file for PulpImport"),
        required=False,
    )
    repo_mapping = EvaluationSerializer(
        help_text=_("Evaluation of proposed 'repo_mapping' file for PulpImport"),
        required=False,
    )


class PulpImportCheckSerializer(ValidateFieldsMixin, serializers.Serializer):
    """
    Check validity of provided import-options.

    Provides the ability to check that an import is 'sane' without having to actually
    create an importer.
    """

    path = serializers.CharField(
        help_text=_("Path to export-tar-gz that will be imported."), required=False
    )
    toc = serializers.CharField(
        help_text=_(
            "Path to a table-of-contents file describing chunks to be validated, "
            "reassembled, and imported."
        ),
        required=False,
    )
    repo_mapping = serializers.CharField(
        help_text=_(
            "Mapping of repo names in an export file to the repo names in Pulp. "
            "For example, if the export has a repo named 'foo' and the repo to "
            "import content into was 'bar', the mapping would be \"{'foo': 'bar'}\"."
        ),
        required=False,
    )

    def validate(self, data):
        data = super().validate(data)
        if "path" not in data and "toc" not in data and "repo_mapping" not in data:
            raise serializers.ValidationError(
                _("One of 'path', 'toc', or 'repo_mapping' must be specified.")
            )
        else:
            return data



# Support plugins dispatching tasks
from pulpcore.tasking.tasks import dispatch  # noqa

from pulpcore.app.tasks import (  # noqa
    fs_publication_export,
    fs_repo_version_export,
    general_create,
    general_create_from_temp_file,
    general_delete,
    general_multi_delete,
    general_update,
    orphan_cleanup,
    reclaim_space,
)
from pulpcore.app.tasks.repository import add_and_remove  # noqa



# Generated by Django 2.2.5 on 2019-09-13 21:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0005_progressreport_code'),
    ]

    operations = [
        migrations.AddField(
            model_name='repository',
            name='plugin_managed',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.11 on 2020-03-27 19:14

import django.core.files.storage
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0023_change_exporter_models'),
    ]

    operations = [
        migrations.AlterField(
            model_name='upload',
            name='file',
            field=models.FileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(location='/var/lib/pulp/upload/'), upload_to=''),
        ),
    ]



# Generated by Django 2.2.17 on 2020-12-05 11:11

import json
import warnings
import subprocess
import tempfile

import gnupg

from django.db import migrations, models


def sign(service, filename):
    completed_process = subprocess.run(
        [service.script, filename], env={}, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    if completed_process.returncode != 0:
        raise RuntimeError()

    try:
        return_value = json.loads(completed_process.stdout)
    except json.JSONDecodeError:
        raise RuntimeError()

    return return_value


def print_warning(service):
    warnings.warn(
        "The public key migration for '{}' failed. The resource will not "
        "be valid. Consider creating a new signing service instead.".format(service.name),
        RuntimeWarning,
    )


def sign_temp_file(service, temp_file):
    try:
        return sign(service, temp_file.name)
    except RuntimeError:
        print_warning(service)
        raise


def update_model_fields(sign_results, service, gpg):
    try:
        with open(sign_results['key'], 'rb') as key:
            public_key = key.read()
            service.public_key = public_key

            import_result = gpg.import_keys(public_key)
            service.pubkey_fingerprint = import_result.fingerprints[0]
            service.save()
    except OSError:
        print_warning(service)


def migrate_public_key_values(apps, schema_editor):
    SigningService = apps.get_model('core', 'SigningService')
    gpg = gnupg.GPG()

    with tempfile.TemporaryDirectory() as temp_directory_name:
        with tempfile.NamedTemporaryFile(dir=temp_directory_name) as temp_file:
            temp_file.write(b'arbitrary data')
            temp_file.flush()

            for service in SigningService.objects.all():
                try:
                    sign_results = sign_temp_file(service, temp_file)
                except RuntimeError:
                    pass
                else:
                    update_model_fields(sign_results, service, gpg)


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0053_remote_headers'),
    ]

    operations = [
        migrations.AddField(
            model_name='signingservice',
            name='public_key',
            field=models.TextField(default=''),
        ),
        migrations.AddField(
            model_name='signingservice',
            name='pubkey_fingerprint',
            field=models.TextField(default=''),
        ),
        migrations.RunPython(migrate_public_key_values, reverse_code=migrations.RunPython.noop),
        migrations.AlterField(
            model_name='signingservice',
            name='public_key',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='signingservice',
            name='pubkey_fingerprint',
            field=models.TextField(),
        ),
    ]



from gettext import gettext as _

from django.shortcuts import get_object_or_404
from django_filters import Filter
from django_filters.rest_framework import DjangoFilterBackend, filters
from drf_spectacular.utils import extend_schema
from rest_framework import mixins, serializers
from rest_framework.filters import OrderingFilter
from rest_framework.decorators import action
from rest_framework.response import Response

from pulpcore.app.models import (
    ContentGuard,
    RBACContentGuard,
    Distribution,
    Publication,
    Content,
)
from pulpcore.app.serializers import (
    ContentGuardSerializer,
    DistributionSerializer,
    PublicationSerializer,
    RBACContentGuardSerializer,
    RBACContentGuardPermissionSerializer,
)
from pulpcore.app.viewsets import (
    AsyncCreateMixin,
    AsyncRemoveMixin,
    AsyncUpdateMixin,
    BaseFilterSet,
    NamedModelViewSet,
)
from pulpcore.app.viewsets.base import DATETIME_FILTER_OPTIONS, NAME_FILTER_OPTIONS
from pulpcore.app.viewsets.custom_filters import (
    IsoDateTimeFilter,
    LabelSelectFilter,
    RepositoryVersionFilter,
)


class PublicationContentFilter(Filter):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("help_text", _("Content Unit referenced by HREF"))
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value is None:
            # user didn't supply a value
            return qs

        if not value:
            raise serializers.ValidationError(detail=_("No value supplied for content filter"))

        # Get the content object from the content_href
        content = NamedModelViewSet.get_resource(value, Content)

        return qs.with_content([content.pk])


class PublicationFilter(BaseFilterSet):
    repository_version = RepositoryVersionFilter()
    pulp_created = IsoDateTimeFilter()
    content = PublicationContentFilter()
    content__in = PublicationContentFilter(field_name="content", lookup_expr="in")

    class Meta:
        model = Publication
        fields = {
            "repository_version": ["exact"],
            "pulp_created": DATETIME_FILTER_OPTIONS,
        }


class ListPublicationViewSet(NamedModelViewSet, mixins.ListModelMixin):
    endpoint_name = "publications"
    queryset = Publication.objects.all()
    serializer_class = PublicationSerializer
    filterset_class = PublicationFilter

    @classmethod
    def is_master_viewset(cls):
        """Do not hide from the routers."""
        return False


class PublicationViewSet(
    NamedModelViewSet, mixins.RetrieveModelMixin, mixins.ListModelMixin, mixins.DestroyModelMixin
):
    endpoint_name = "publications"
    queryset = Publication.objects.exclude(complete=False)
    serializer_class = PublicationSerializer
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    filterset_class = PublicationFilter
    ordering = ("-pulp_created",)


class ContentGuardFilter(BaseFilterSet):
    name = filters.CharFilter()

    class Meta:
        model = ContentGuard
        fields = {
            "name": NAME_FILTER_OPTIONS,
        }


class BaseContentGuardViewSet(NamedModelViewSet):
    endpoint_name = "contentguards"
    serializer_class = ContentGuardSerializer
    queryset = ContentGuard.objects.all()
    filterset_class = ContentGuardFilter


class ListContentGuardViewSet(
    BaseContentGuardViewSet,
    mixins.ListModelMixin,
):
    """Endpoint to list all contentguards."""

    @classmethod
    def is_master_viewset(cls):
        """Do not hide from the routers."""
        return False


class ContentGuardViewSet(
    BaseContentGuardViewSet,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
):
    """
    A viewset for contentguards.
    """


class RBACContentGuardViewSet(ContentGuardViewSet):
    """
    Viewset for creating contentguards that use RBAC to protect content.
    Has add and remove actions for managing permission for users and groups to download content
    protected by this guard.
    """

    endpoint_name = "rbac"
    serializer_class = RBACContentGuardSerializer
    queryset = RBACContentGuard.objects.all()

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_perms:core.add_rbaccontentguard",
            },
            {
                "action": ["retrieve"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.view_rbaccontentguard",
            },
            {
                "action": ["update", "partial_update", "assign_permission", "remove_permission"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.change_rbaccontentguard",
            },
            {
                "action": ["destroy"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.delete_rbaccontentguard",
            },
            {
                "action": ["download"],  # This is the action for the content guard permit
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.download_rbaccontentguard",
            },
        ],
        "permissions_assignment": [
            {
                "function": "add_for_object_creator",
                "parameters": None,
                "permissions": [
                    "core.add_rbaccontentguard",
                    "core.view_rbaccontentguard",
                    "core.change_rbaccontentguard",
                    "core.delete_rbaccontentguard",
                    "core.download_rbaccontentguard",
                ],
            },
        ],
    }

    @extend_schema(summary="Add download permission", responses={201: RBACContentGuardSerializer})
    @action(methods=["post"], detail=True, serializer_class=RBACContentGuardPermissionSerializer)
    def assign_permission(self, request, pk):
        """Give users and groups the `download` permission"""
        guard = get_object_or_404(RBACContentGuard, pk=pk)
        names = self.get_serializer(data=request.data)
        names.is_valid(raise_exception=True)
        guard.add_can_download(users=names.data["usernames"], groups=names.data["groupnames"])
        self.serializer_class = RBACContentGuardSerializer
        serializer = self.get_serializer(guard)
        return Response(serializer.data, status=201)

    @extend_schema(
        summary="Remove download permission", responses={201: RBACContentGuardSerializer}
    )
    @action(methods=["post"], detail=True, serializer_class=RBACContentGuardPermissionSerializer)
    def remove_permission(self, request, pk):
        """Remove `download` permission from users and groups"""
        guard = get_object_or_404(RBACContentGuard, pk=pk)
        names = self.get_serializer(data=request.data)
        names.is_valid(raise_exception=True)
        guard.remove_can_download(users=names.data["usernames"], groups=names.data["groupnames"])
        self.serializer_class = RBACContentGuardSerializer
        serializer = self.get_serializer(guard)
        return Response(serializer.data, status=201)


class DistributionFilter(BaseFilterSet):
    # e.g.
    # /?name=foo
    # /?name__in=foo,bar
    # /?base_path__contains=foo
    # /?base_path__icontains=foo
    name = filters.CharFilter()
    base_path = filters.CharFilter()
    pulp_label_select = LabelSelectFilter()

    class Meta:
        model = Distribution
        fields = {
            "name": NAME_FILTER_OPTIONS,
            "base_path": ["exact", "contains", "icontains", "in"],
        }


class DistributionViewSet(
    NamedModelViewSet,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    AsyncCreateMixin,
    AsyncRemoveMixin,
    AsyncUpdateMixin,
):
    """
    Provides read and list methods and also provides asynchronous CUD methods to dispatch tasks
    with reservation that lock all Distributions preventing race conditions during base_path
    checking.
    """

    endpoint_name = "distributions"
    queryset = Distribution.objects.all()
    serializer_class = DistributionSerializer
    filterset_class = DistributionFilter

    def async_reserved_resources(self, instance):
        """Return resource that locks all Distributions."""
        return ["/api/v3/distributions/"]



# Generated by Django 2.2.11 on 2020-04-13 16:39

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0027_export_backend'),
    ]

    operations = [
        migrations.CreateModel(
            name='Import',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('params', django.contrib.postgres.fields.jsonb.JSONField(null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Importer',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('pulp_type', models.TextField(default=None)),
                ('name', models.TextField(db_index=True, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PulpImport',
            fields=[
                ('import_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='core_pulp_export', serialize=False, to='core.Import')),
            ],
            options={
                'default_related_name': '%(app_label)s_pulp_export',
            },
            bases=('core.import',),
        ),
        migrations.CreateModel(
            name='PulpImporter',
            fields=[
                ('importer_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='core_pulp_importer', serialize=False, to='core.Importer')),
            ],
            options={
                'default_related_name': '%(app_label)s_pulp_importer',
            },
            bases=('core.importer',),
        ),
        migrations.AddField(
            model_name='import',
            name='importer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Importer'),
        ),
        migrations.AddField(
            model_name='import',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='core.Task'),
        ),
        migrations.CreateModel(
            name='PulpImporterRepository',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('source_repo', models.TextField()),
                ('repository', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Repository')),
                ('pulp_importer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='repo_map', to='core.PulpImporter')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# Generated by Django 2.2.11 on 2020-04-21 17:12

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0028_import_importer_pulpimporter_pulpimporterrepository'),
    ]

    operations = [
        migrations.AlterField(
            model_name='export',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Task'),
        ),
    ]



from drf_spectacular.utils import extend_schema

from django.db import DatabaseError
from django.db.utils import IntegrityError

from pulpcore.app import tasks
from pulpcore.plugin.serializers import (
    ArtifactSerializer,
    AsyncOperationResponseSerializer,
)
from pulpcore.plugin.models import Artifact, PulpTemporaryFile
from pulpcore.plugin.tasking import dispatch
from pulpcore.plugin.viewsets import (
    ContentViewSet,
    OperationPostponedResponse,
)


class DefaultDeferredContextMixin:
    """A mixin that provides a method for retrieving the default deferred context."""

    def get_deferred_context(self, request):
        """
        Supply context for deferred validation.

        When overwriting this method, it must return a dict, that is JSON serializable by
        and does _not_ contain 'request' as a key.
        """
        return {}


class NoArtifactContentUploadViewSet(DefaultDeferredContextMixin, ContentViewSet):
    """A ViewSet for uploads that do not require to store an uploaded content as an Artifact."""

    @extend_schema(
        description="Trigger an asynchronous task to create content,"
        "optionally create new repository version.",
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request):
        """Create a content unit."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        task_payload = {k: v for k, v in request.data.items()}
        file_content = task_payload.pop("file", None)

        temp_file = PulpTemporaryFile.init_and_validate(file_content)
        temp_file.save()

        resources = []
        repository = serializer.validated_data.get("repository")
        if repository:
            resources.append(repository)

        app_label = self.queryset.model._meta.app_label
        task = dispatch(
            tasks.base.general_create_from_temp_file,
            exclusive_resources=resources,
            args=(app_label, serializer.__class__.__name__, str(temp_file.pk)),
            kwargs={"data": task_payload, "context": self.get_deferred_context(request)},
        )
        return OperationPostponedResponse(task, request)


class SingleArtifactContentUploadViewSet(DefaultDeferredContextMixin, ContentViewSet):
    """A ViewSet which can be used to store an uploaded content as an Artifact."""

    @extend_schema(
        description="Trigger an asynchronous task to create content,"
        "optionally create new repository version.",
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request):
        """Create a content unit."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        task_payload = self.init_content_data(serializer, request)

        resources = []
        repository = serializer.validated_data.get("repository")
        if repository:
            resources.append(repository)

        app_label = self.queryset.model._meta.app_label
        task = dispatch(
            tasks.base.general_create,
            args=(app_label, serializer.__class__.__name__),
            exclusive_resources=resources,
            kwargs={
                "data": task_payload,
                "context": self.get_deferred_context(request),
            },
        )
        return OperationPostponedResponse(task, request)

    def init_content_data(self, serializer, request):
        """Initialize the reference to an Artifact along with relevant task's payload data."""
        artifact = serializer.validated_data["artifact"]

        task_payload = {k: v for k, v in request.data.items()}
        if task_payload.pop("file", None):
            # in the upload code path make sure, the artifact exists, and the 'file'
            # parameter is replaced by 'artifact'
            try:
                artifact.save()
            except IntegrityError:
                # if artifact already exists, let's use it
                try:
                    artifact = Artifact.objects.get(sha256=artifact.sha256)
                    artifact.touch()
                except (Artifact.DoesNotExist, DatabaseError):
                    # the artifact has since been removed from when we first attempted to save it
                    artifact.save()

            task_payload["artifact"] = ArtifactSerializer(
                artifact, context={"request": request}
            ).data["pulp_href"]

        return task_payload



# Generated by Django 2.2.7 on 2019-11-12 20:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0015_auto_20191112_1426'),
    ]

    operations = [
        migrations.AlterField(
            model_name='basedistribution',
            name='base_path',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='basedistribution',
            name='name',
            field=models.TextField(db_index=True, unique=True),
        ),
        migrations.AlterField(
            model_name='contentappstatus',
            name='name',
            field=models.TextField(db_index=True, unique=True),
        ),
        migrations.AlterField(
            model_name='contentguard',
            name='name',
            field=models.TextField(db_index=True, unique=True),
        ),
        migrations.AlterField(
            model_name='filesystemexporter',
            name='name',
            field=models.TextField(db_index=True, unique=True),
        ),
        migrations.AlterField(
            model_name='remote',
            name='name',
            field=models.TextField(db_index=True, unique=True),
        ),
        migrations.AlterField(
            model_name='repository',
            name='name',
            field=models.TextField(db_index=True, unique=True),
        ),
        migrations.AlterField(
            model_name='reservedresource',
            name='resource',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='reservedresourcerecord',
            name='resource',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='worker',
            name='name',
            field=models.TextField(db_index=True, unique=True),
        ),
    ]



# Generated by Django 3.2.5 on 2021-07-21 17:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0068_add_timestamp_of_interest'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accesspolicy',
            name='permissions_assignment',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='accesspolicy',
            name='statements',
            field=models.JSONField(),
        ),
        migrations.AlterField(
            model_name='export',
            name='params',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='import',
            name='params',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='pulpexport',
            name='output_file_info',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='pulpexport',
            name='toc_info',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='remote',
            name='headers',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='args',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='error',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='kwargs',
            field=models.JSONField(null=True),
        ),
    ]



# Generated by Django 2.2.14 on 2020-08-25 13:40

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0042_rbac_for_tasks'),
    ]

    operations = [
        migrations.AddField(
            model_name='pulpexport',
            name='toc_info',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
    ]



from gettext import gettext as _

from rest_framework import serializers
from rest_framework.validators import UniqueValidator

from pulpcore.app import models
from pulpcore.app.serializers import IdentityField, ModelSerializer


class AccessPolicySerializer(ModelSerializer):
    """Serializer for AccessPolicy."""

    pulp_href = IdentityField(view_name="access_policies-detail")

    permissions_assignment = serializers.ListField(
        child=serializers.DictField(),
        help_text=_(
            "List of callables that define the new permissions to be created for new objects."
        ),
    )

    statements = serializers.ListField(
        child=serializers.DictField(),
        help_text=_("List of policy statements defining the policy."),
    )

    viewset_name = serializers.CharField(
        help_text=_("The name of ViewSet this AccessPolicy authorizes."),
        validators=[UniqueValidator(queryset=models.AccessPolicy.objects.all())],
        read_only=True,
    )

    customized = serializers.BooleanField(
        help_text=_("True if the AccessPolicy has been user-modified. False otherwise."),
        read_only=True,
    )

    class Meta:
        model = models.AccessPolicy
        fields = ModelSerializer.Meta.fields + (
            "permissions_assignment",
            "statements",
            "viewset_name",
            "customized",
        )

    def validate(self, data):
        """ "
        Validate the AccessPolicy.

        This ensures that the customized boolean will be set to True anytime the user modifies it.
        """
        data = super().validate(data)
        if data:
            data["customized"] = True
        return data



import os
import re
from urllib.parse import urljoin

from django.core.exceptions import FieldDoesNotExist
from django.utils.html import strip_tags
from drf_spectacular.drainage import reset_generator_stats
from drf_spectacular.generators import SchemaGenerator
from drf_spectacular.openapi import AutoSchema
from drf_spectacular.plumbing import (
    ResolvedComponent,
    build_basic_type,
    build_parameter_type,
    build_root_object,
    force_instance,
    normalize_result_object,
    resolve_django_path_parameter,
    resolve_regex_path_parameter,
)
from drf_spectacular.settings import spectacular_settings
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter
from rest_framework import mixins, serializers
from rest_framework.schemas.utils import get_pk_description

from pulpcore.app.apps import pulp_plugin_configs


class PulpAutoSchema(AutoSchema):
    """Pulp Auto Schema."""

    method_mapping = {
        "get": "read",
        "post": "create",
        "put": "update",
        "patch": "partial_update",
        "delete": "delete",
    }

    def _tokenize_path(self):
        """
        Tokenize path.

        drf_spectacular uses this to tokenize the path:
            "/my/path/to/{variable}/api" => ["my", "path", "to", "api"]

        But pulp manipulates the path:
            "/pulp/api/v3/artifacts/{pulp_id}" == "{artifact_href}"

        This method extends _tokenize_path to handle pulp cases.

        """
        tokenized_path = []

        if getattr(self.view, "parent_viewset", None):
            tokenized_path.extend(self.view.parent_viewset.endpoint_pieces())

        if getattr(self.view, "endpoint_pieces", None):
            tokenized_path.extend(self.view.endpoint_pieces())

        if not tokenized_path:
            tokenized_path = super()._tokenize_path()
            if not tokenized_path and getattr(self.view, "get_view_name", None):
                tokenized_path.extend(self.view.get_view_name().split())

        path = "/".join(tokenized_path).replace("pulp/api/v3/", "")
        tokenized_path = path.split("/")

        return tokenized_path

    def get_tags(self):
        """
        Generate tags.

        For bindings, tags are used to group operation ids into same class.
        Example:
            "Content: Files" => ContentFilesAPI

        The path is used to generate a tag:
            "/pulp/api/v3/content/file/files/" => "Content: Files"

        For customize the tag, please set `pulp_tag_name` at your view.
        Example:
            class MyViewSet(ViewSet):
                pulp_tag_name = "Pulp: Customized Tag"

        """
        pulp_tag_name = getattr(self.view, "pulp_tag_name", False)
        if pulp_tag_name:
            return [pulp_tag_name]

        tokenized_path = self._tokenize_path()

        subpath = "/".join(tokenized_path)
        operation_keys = subpath.replace("pulp/api/v3/", "").split("/")
        operation_keys = [i.title() for i in operation_keys]
        tags = operation_keys
        if len(operation_keys) > 2:
            del operation_keys[1]
        if len(operation_keys) > 1:
            operation_keys[0] = "{key}:".format(key=operation_keys[0])
        tags = [" ".join(operation_keys)]

        return tags

    def get_operation_id_action(self):
        """
        Get action from operation_id.

        - For default actions: maps methods to action
            "patch" => "partial_update"
        - For customized actions: return the customized action
            e.g. "sync"
        """
        action_name = getattr(self.view, "action", self.method.lower())
        if action_name not in ["retrieve", "list", "destroy", "create"]:
            return action_name

        if self.method.lower() == "get" and self._is_list_view():
            return "list"

        return self.method_mapping[self.method.lower()]

    def get_operation_id(self):
        """
        Get operation id.

        Combines tokenized_path with action.
        Example:
            path = "/my/path/to/{variable}/api"
            method = "patch"

            Return "my_path_to_api_partial_update"

        """
        tokenized_path = self._tokenize_path()
        tokenized_path = [t.replace("-", "_").replace("/", "_").lower() for t in tokenized_path]

        return "_".join(tokenized_path + [self.get_operation_id_action()])

    def get_summary(self):
        """
        Returns summary of operation.
        This is the value that is displayed in the ReDoc document as the short name for the API
        operation.
        """
        if not hasattr(self.view, "queryset") or self.view.queryset is None:
            return None
        model = self.view.queryset.model
        operation = self.get_operation_id().split("_")[-1]
        resource = model._meta.verbose_name
        article = "a"
        if resource[0].lower() in "aeiou":
            article = "an"
        if operation == "read":
            return f"Inspect {article} {resource}"
        if operation == "list":
            resource = model._meta.verbose_name_plural
            return f"List {resource}"
        if operation == "create":
            return f"Create {article} {resource}"
        if operation == "update":
            return f"Update {article} {resource}"
        if operation == "delete":
            return f"Delete {article} {resource}"
        if operation == "partial_update":
            return f"Partially update {article} {resource}"

    def _get_serializer_name(self, serializer, direction):
        """
        Get serializer name.
        """
        name = super()._get_serializer_name(serializer, direction)
        if direction == "request":
            name = name[:-7]
        elif direction == "response" and "Response" not in name:
            name = name + "Response"
        return name

    def map_parsers(self):
        """
        Get request parsers.

        Handling cases with `FileField`.
        """
        parsers = super().map_parsers()
        serializer = force_instance(self.get_request_serializer())
        for field_name, field in getattr(serializer, "fields", {}).items():
            if isinstance(field, serializers.FileField) and self.method in ("PUT", "PATCH", "POST"):
                return ["multipart/form-data", "application/x-www-form-urlencoded"]
        return parsers

    def _get_request_body(self):
        """Get request body."""
        request_body = super()._get_request_body()
        if request_body:
            request_body["required"] = True
        return request_body

    def _resolve_path_parameters(self, variables):
        """
        Resolve path parameters.

        Extended to omit undesired warns.
        """
        model = getattr(getattr(self.view, "queryset", None), "model", None)
        parameters = []

        for variable in variables:
            schema = build_basic_type(OpenApiTypes.STR)
            description = ""

            resolved_parameter = resolve_django_path_parameter(
                self.path_regex,
                variable,
                self.map_renderers("format"),
            )
            if not resolved_parameter:
                resolved_parameter = resolve_regex_path_parameter(self.path_regex, variable)

            if resolved_parameter:
                schema = resolved_parameter["schema"]
            elif model:
                try:
                    model_field = model._meta.get_field(variable)
                    schema = self._map_model_field(model_field, direction=None)
                    # strip irrelevant meta data
                    irrelevant_field_meta = ["readOnly", "writeOnly", "nullable", "default"]
                    schema = {k: v for k, v in schema.items() if k not in irrelevant_field_meta}
                    if "description" not in schema and model_field.primary_key:
                        description = get_pk_description(model, model_field)
                except FieldDoesNotExist:
                    pass

            parameters.append(
                build_parameter_type(
                    name=variable,
                    location=OpenApiParameter.PATH,
                    description=description,
                    schema=schema,
                )
            )

        return parameters

    def resolve_serializer(self, serializer, direction):
        """Serializer to component."""
        component_schema = self._map_serializer(serializer, direction)
        if not component_schema.get("properties", {}):
            component = ResolvedComponent(
                name=self._get_serializer_name(serializer, direction),
                type=ResolvedComponent.SCHEMA,
                object=serializer,
            )
            if component in self.registry:
                return self.registry[component]

            component.schema = component_schema
            self.registry.register(component)

        else:
            component = super().resolve_serializer(serializer, direction)

        return component

    def _get_response_bodies(self):
        """
        Handle response status code.
        """
        response = super()._get_response_bodies()
        if (
            self.method == "POST"
            and issubclass(self.view.__class__, mixins.CreateModelMixin)
            and "200" in response
        ):
            response["201"] = response.pop("200")

        return response


class PulpSchemaGenerator(SchemaGenerator):
    """Pulp Schema Generator."""

    @staticmethod
    def get_parameter_slug_from_model(model, prefix, pulp_model_alias=None):
        """Returns a path parameter name for the resource associated with the model.
        Args:
            model (django.db.models.Model): The model for which a path parameter name is needed
            prefix (str): Optional prefix to add to the slug
            pulp_model_alias (str): Optional model name to use instead of model.__name__
        Returns:
            str: *pulp_href where * is the model name in all lower case letters
        """
        app_label = model._meta.app_label
        model_name = pulp_model_alias or model.__name__
        parts = [part.lower() for part in re.findall("[A-Z][^A-Z]*", model_name)]
        if prefix:
            parts.insert(0, prefix)
        elif app_label != "core":
            parts.insert(0, app_label)
        parts.append("href")
        return "_".join(parts)

    @staticmethod
    def get_pk_path_param_name_from_model(model):
        """Returns a specific name for the primary key of a model.

        Args:
            model (django.db.models.Model): The model for which a path parameter name is needed

        Returns:
            str: *_pk where * is the model name in all lower case letters
        """
        return "%s_pk" % "_".join(
            [part.lower() for part in re.findall("[A-Z][^A-Z]*", model.__name__)]
        )

    def convert_endpoint_path_params(self, path, view, schema):
        """Replaces all 'pulp_id' path parameters with a specific name for the primary key.
        This method is used to ensure that the primary key name is consistent between nested
        endpoints. get_endpoints() returns paths that use 'pulp_id' for the top level path and a
        specific name for the nested paths. e.g.: repository_pk.
        This ensures that when the endpoints are sorted, the parent endpoint appears before the
        endpoints nested under it.
        Returns:
            path(str): The modified path.
        """
        if "{" not in path:
            return path
        if getattr(view, "queryset", None) is None:
            if hasattr(view, "model"):
                resource_model = view.model
            else:
                return path
        else:
            resource_model = view.queryset.model
        if resource_model:
            prefix = None
            pulp_model_alias = getattr(view, "pulp_model_alias", None)
            parent_viewset = getattr(view, "parent_viewset", None)
            if parent_viewset:
                if view.action in ["list", "create"]:
                    resource_model = parent_viewset.queryset.model
                    pulp_model_alias = getattr(parent_viewset, "pulp_model_alias", None)

                else:
                    parent_app_label = parent_viewset.queryset.model._meta.app_label
                    prefix_parts = [parent_viewset.endpoint_name]
                    if parent_app_label != "core":
                        prefix_parts.insert(0, parent_app_label)
                    prefix = "_".join(prefix_parts).replace("-", "_").replace("/", "_").lower()
            param_name = self.get_parameter_slug_from_model(
                resource_model, prefix, pulp_model_alias
            )
            resource_path = "%s}/" % path.rsplit(sep="}", maxsplit=1)[0]
            path = path.replace(resource_path, "{%s}" % param_name)
        return path

    def parse(self, input_request, public):
        """Iterate endpoints generating per method path operations."""
        result = {}
        self._initialise_endpoints()
        endpoints = self._get_paths_and_endpoints()

        if spectacular_settings.SCHEMA_PATH_PREFIX is None:
            # estimate common path prefix if none was given. only use it if we encountered more
            # than one view to prevent emission of erroneous and unnecessary fallback names.
            non_trivial_prefix = len(set([view.__class__ for _, _, _, view in endpoints])) > 1
            if non_trivial_prefix:
                path_prefix = os.path.commonpath([path for path, _, _, _ in endpoints])
                path_prefix = re.escape(path_prefix)  # guard for RE special chars in path
            else:
                path_prefix = "/"
        else:
            path_prefix = spectacular_settings.SCHEMA_PATH_PREFIX
        if not path_prefix.startswith("^"):
            path_prefix = "^" + path_prefix  # make sure regex only matches from the start

        # Adding plugin filter
        plugins = None
        # /pulp/api/v3/docs/api.json?plugin=pulp_file
        if input_request and "plugin" in input_request.query_params:
            plugins = [input_request.query_params["plugin"]]

        for path, path_regex, method, view in endpoints:
            plugin = view.__module__.split(".")[0]
            if plugins and plugin not in plugins:  # plugin filter
                continue

            view.request = spectacular_settings.GET_MOCK_REQUEST(method, path, view, input_request)

            if not (public or self.has_view_permissions(path, method, view)):
                continue

            schema = view.schema

            path = self.convert_endpoint_path_params(path, view, schema)

            # beware that every access to schema yields a fresh object (descriptor pattern)
            operation = schema.get_operation(path, path_regex, path_prefix, method, self.registry)

            # operation was manually removed via @extend_schema
            if not operation:
                continue

            # Removes html tags from OpenAPI schema
            if input_request is None or "include_html" not in input_request.query_params:
                if "description" in operation:
                    operation["description"] = strip_tags(operation["description"])

            # operationId as actions [list, read, sync, modify, create, delete, ...]
            if input_request and "bindings" in input_request.query_params:
                tokenized_path = schema._tokenize_path()
                tokenized_path = "_".join(
                    [t.replace("-", "_").replace("/", "_").lower() for t in tokenized_path]
                )
                action = schema.get_operation_id_action()
                if f"{tokenized_path}_{action}" == operation["operationId"]:
                    operation["operationId"] = action

            # Adding query parameters
            if "parameters" in operation and schema.method.lower() == "get":
                fields_paramenter = build_parameter_type(
                    name="fields",
                    schema={"type": "string"},
                    location=OpenApiParameter.QUERY,
                    description="A list of fields to include in the response.",
                )
                operation["parameters"].append(fields_paramenter)
                not_fields_paramenter = build_parameter_type(
                    name="exclude_fields",
                    schema={"type": "string"},
                    location=OpenApiParameter.QUERY,
                    description="A list of fields to exclude from the response.",
                )
                operation["parameters"].append(not_fields_paramenter)

            # Normalise path for any provided mount url.
            if path.startswith("/"):
                path = path[1:]

            if not path.startswith("{"):
                path = urljoin(self.url or "/", path)

            result.setdefault(path, {})
            result[path][method.lower()] = operation

        return result

    def get_schema(self, request=None, public=False):
        """Generate a OpenAPI schema."""
        reset_generator_stats()
        result = build_root_object(
            paths=self.parse(request, public),
            components=self.registry.build(spectacular_settings.APPEND_COMPONENTS),
            version=self.api_version or getattr(request, "version", None),
        )
        for hook in spectacular_settings.POSTPROCESSING_HOOKS:
            result = hook(result=result, generator=self, request=request, public=public)

        # Basically I'm doing it to get pulp logo at redoc page
        result["info"]["x-logo"] = {
            "url": "https://pulp.plan.io/attachments/download/517478/pulp_logo_word_rectangle.svg"
        }

        # Adding plugin version config
        result["info"]["x-pulp-app-versions"] = {}
        for app in pulp_plugin_configs():
            result["info"]["x-pulp-app-versions"][app.label] = app.version

        # Adding current host as server (it will provide a default value for the bindings)
        server_url = "http://localhost:24817" if not request else request.build_absolute_uri("/")
        result["servers"] = [{"url": server_url}]

        return normalize_result_object(result)



# Generated by Django 2.2.3 on 2019-08-14 12:15

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0003_remove_upload_completed'),
    ]

    operations = [
        migrations.CreateModel(
            name='ReservedResourceRecord',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('resource', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TaskReservedResourceRecord',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('resource', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.ReservedResourceRecord')),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Task')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='reservedresourcerecord',
            name='tasks',
            field=models.ManyToManyField(related_name='reserved_resources_record', through='core.TaskReservedResourceRecord', to='core.Task'),
        ),
    ]



"""Tests related to content delivery."""
import hashlib
import unittest
from random import choice
from urllib.parse import urljoin

from pulp_smash import api, config, utils
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task, PulpTestCase
from pulp_smash.pulp3.constants import ON_DEMAND_DOWNLOAD_POLICIES
from pulp_smash.pulp3.utils import (
    download_content_unit,
    gen_distribution,
    gen_repo,
    get_content,
    sync,
)
from requests import HTTPError

from pulpcore.client.pulp_file import (
    PublicationsFileApi,
    RemotesFileApi,
    RepositoriesFileApi,
    RepositorySyncURL,
    DistributionsFileApi,
)
from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_CONTENT_NAME,
    FILE_DISTRIBUTION_PATH,
    FILE_FIXTURE_URL,
    FILE_FIXTURE_MANIFEST_URL,
    FILE_FIXTURE_WITH_MISSING_FILES_MANIFEST_URL,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
)
from pulpcore.tests.functional.api.using_plugin.utils import (
    create_file_publication,
    gen_file_remote,
    gen_file_client,
)
from pulpcore.tests.functional.api.using_plugin.utils import (  # noqa:F401
    set_up_module as setUpModule,
)


class ContentDeliveryTestCase(unittest.TestCase):
    """Content delivery breaks when delete remote - lazy download policy.

    Deleting a remote that was used in a sync with either the on_demand or
    streamed options can break published data. Specifically, clients who want
    to fetch content that a remote was providing access to would begin to
    404. Recreating a remote and re-triggering a sync will cause these broken
    units to recover again.

    This test targets the following issue:

    * `Pulp #4464 <https://pulp.plan.io/issues/4464>`_
    """

    def test_content_remote_delete(self):
        """Assert that an HTTP error is raised when remote is deleted.

        Also verify that the content can be downloaded from Pulp once the
        remote is recreated and another sync is triggered.
        """
        cfg = config.get_config()
        delete_orphans()
        client = api.Client(cfg, api.page_handler)

        repo = client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["pulp_href"])

        body = gen_file_remote(policy=choice(ON_DEMAND_DOWNLOAD_POLICIES))
        remote = client.post(FILE_REMOTE_PATH, body)

        # Sync the repository using a lazy download policy.
        sync(cfg, remote, repo)
        repo = client.get(repo["pulp_href"])

        publication = create_file_publication(cfg, repo)
        self.addCleanup(client.delete, publication["pulp_href"])

        # Delete the remote.
        client.delete(remote["pulp_href"])

        body = gen_distribution()
        body["publication"] = publication["pulp_href"]
        distribution = client.using_handler(api.task_handler).post(FILE_DISTRIBUTION_PATH, body)
        self.addCleanup(client.delete, distribution["pulp_href"])

        unit_path = choice(
            [content_unit["relative_path"] for content_unit in get_content(repo)[FILE_CONTENT_NAME]]
        )

        # Assert that an HTTP error is raised when one to fetch content from
        # the distribution once the remote was removed.
        with self.assertRaises(HTTPError) as ctx:
            download_content_unit(cfg, distribution, unit_path)
        for key in ("not", "found"):
            self.assertIn(key, ctx.exception.response.reason.lower())

        # Recreating a remote and re-triggering a sync will cause these broken
        # units to recover again.
        body = gen_file_remote(policy=choice(ON_DEMAND_DOWNLOAD_POLICIES))
        remote = client.post(FILE_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote["pulp_href"])

        sync(cfg, remote, repo)
        repo = client.get(repo["pulp_href"])

        content = download_content_unit(cfg, distribution, unit_path)
        pulp_hash = hashlib.sha256(content).hexdigest()

        fixtures_hash = hashlib.sha256(
            utils.http_get(urljoin(FILE_FIXTURE_URL, unit_path))
        ).hexdigest()

        self.assertEqual(pulp_hash, fixtures_hash)


class RemoteArtifactUpdateTestCase(PulpTestCase):
    @classmethod
    def setUpClass(cls):
        """Clean out Pulp before testing."""
        delete_orphans()
        client = gen_file_client()
        cls.repo_api = RepositoriesFileApi(client)
        cls.remote_api = RemotesFileApi(client)
        cls.publication_api = PublicationsFileApi(client)
        cls.distributions_api = DistributionsFileApi(client)
        cls.cfg = config.get_config()

    def tearDown(self):
        """Clean up Pulp after testing."""
        self.doCleanups()
        delete_orphans()

    def test_remote_artifact_url_update(self):
        """Test that downloading on_demand content works after a repository layout change."""

        FILE_NAME = "1.iso"

        # 1. Create a remote, repository and distribution - remote URL has links that should 404
        remote_config = gen_file_remote(
            policy="on_demand", url=FILE_FIXTURE_WITH_MISSING_FILES_MANIFEST_URL
        )
        remote = self.remote_api.create(remote_config)
        self.addCleanup(self.remote_api.delete, remote.pulp_href)

        repo = self.repo_api.create(gen_repo(autopublish=True, remote=remote.pulp_href))
        self.addCleanup(self.repo_api.delete, repo.pulp_href)

        body = gen_distribution(repository=repo.pulp_href)
        distribution_response = self.distributions_api.create(body)
        created_resources = monitor_task(distribution_response.task).created_resources
        distribution = self.distributions_api.read(created_resources[0])
        self.addCleanup(self.distributions_api.delete, distribution.pulp_href)

        # 2. Sync the repository, verify that downloading artifacts fails
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)

        sync_response = self.repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)

        with self.assertRaises(HTTPError):
            download_content_unit(self.cfg, distribution.to_dict(), FILE_NAME)

        # 3. Update the remote URL with one that works, sync again, check that downloading
        # artifacts works.
        update_response = self.remote_api.update(
            remote.pulp_href, gen_file_remote(policy="on_demand", url=FILE_FIXTURE_MANIFEST_URL)
        )
        monitor_task(update_response.task)

        sync_response = self.repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)

        content = download_content_unit(self.cfg, distribution.to_dict(), FILE_NAME)
        pulp_hash = hashlib.sha256(content).hexdigest()

        fixtures_hash = hashlib.sha256(
            utils.http_get(urljoin(FILE_FIXTURE_URL, FILE_NAME))
        ).hexdigest()

        self.assertEqual(pulp_hash, fixtures_hash)



from .base import (  # noqa
    AdvisoryLockError,
    PulpException,
    ResourceImmutableError,
    TimeoutException,
    exception_to_dict,
)
from .http import MissingResource  # noqa
from .validation import (  # noqa
    DigestValidationError,
    SizeValidationError,
    ValidationError,
    MissingDigestValidationError,
    UnsupportedDigestValidationError,
)



from drf_spectacular.utils import extend_schema
from rest_framework.viewsets import ViewSet

from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.serializers import AsyncOperationResponseSerializer, ReclaimSpaceSerializer
from pulpcore.app.tasks import reclaim_space
from pulpcore.tasking.tasks import dispatch


class ReclaimSpaceViewSet(ViewSet):
    """
    Viewset for reclaim disk space endpoint.
    """

    serializer_class = ReclaimSpaceSerializer

    @extend_schema(
        description="Trigger an asynchronous space reclaim operation.",
        responses={202: AsyncOperationResponseSerializer},
    )
    def reclaim(self, request):
        """
        Triggers an asynchronous space reclaim operation.
        """
        serializer = ReclaimSpaceSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        repos = serializer.validated_data.get("repo_hrefs", [])
        keeplist = serializer.validated_data.get("repo_versions_keeplist", [])
        reclaim_repo_pks = []
        keeplist_rv_pks = []
        for repo in repos:
            reclaim_repo_pks.append(repo.pk)
        for rv in keeplist:
            repos.append(rv.repository)
            keeplist_rv_pks.append(rv.pk)

        task = dispatch(
            reclaim_space,
            shared_resources=repos,
            kwargs={
                "repo_pks": reclaim_repo_pks,
                "keeplist_rv_pks": keeplist_rv_pks,
            },
        )

        return OperationPostponedResponse(task, request)



from getpass import getpass
from gettext import gettext as _

from django.contrib.auth import get_user_model
from django.core.management import BaseCommand, CommandError

User = get_user_model()


class Command(BaseCommand):
    """
    Django management command for resetting the password of the 'admin' user in Pulp.
    """

    help = _('Resets "admin" user\'s password.')

    def add_arguments(self, parser):
        exclusive = parser.add_mutually_exclusive_group()
        exclusive.add_argument(
            "--random",
            action="store_true",
            dest="random",
            default=False,
            help=_("Generate random password for 'admin' user."),
        )
        exclusive.add_argument(
            "-p",
            "--password",
            dest="password",
            default=None,
            help=_("INSECURE: Use the given password for 'admin' user."),
        )

    def handle(self, *args, **options):
        user = User.objects.get_or_create(username="admin", is_superuser=True, is_staff=True)[0]
        if options["random"]:
            password = User.objects.make_random_password(length=20)
            user.set_password(password)
            user.save()
            self.stdout.write(_('Successfully set "admin" user\'s password to "%s".') % password)
        else:
            if options["password"]:
                password = options["password"]
            else:
                # this bit duplicates behavior in the builtin "changepassword" command, so
                # at this point we can probably just call out to that command, leaving this one
                # focused on generating random passwords or setting passwords in automation
                password = getpass(_('Please enter new password for user "admin": '))
                password2 = getpass(_('Please enter new password for user "admin" again: '))
                if password != password2:
                    raise CommandError(_("The passwords did not match."))

            if not password:
                raise CommandError(_("The password must be at least 1 character long."))

            user.set_password(password)
            user.save()
            self.stdout.write(_('Successfully set password for "admin" user.'))



import click
import logging
import os
import select

import django

from pulpcore.app.loggers import deprecation_logger

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "pulpcore.app.settings")
# Until Django supports async ORM natively this is the best we can do given these parts of Pulp
# run in coroutines. We try to ensure it is safe by never passing ORM data between co-routines.
os.environ.setdefault("DJANGO_ALLOW_ASYNC_UNSAFE", "true")

django.setup()

from pulpcore.tasking.pulpcore_worker import NewPulpWorker  # noqa: E402: module level not at top


_logger = logging.getLogger(__name__)


@click.option(
    "--resource-manager",
    is_flag=True,
    help="Whether this worker should be started as a resource-manager",
)
@click.option("--pid", help="Write the process ID number to a file at the specified path")
@click.command()
def worker(resource_manager, pid):
    """A Pulp worker."""

    if pid:
        with open(os.path.expanduser(pid), "w") as fp:
            fp.write(str(os.getpid()))

    if resource_manager:
        _logger.warn("Attempting to start a resource-manager with the distributed tasking system")
        deprecation_logger.warn(
            "The `--resource-manager` option of the pulpcore-worker entrypoint is deprecated and"
            " will be removed in pulpcore 3.17."
        )
        select.select([], [], [])
    _logger.info("Starting distributed type worker")

    NewPulpWorker().run_forever()



from gettext import gettext as _

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.urls import reverse
from guardian.models.models import GroupObjectPermission
from rest_framework import serializers

from pulpcore.app.serializers import IdentityField, ValidateFieldsMixin
from pulpcore.app.util import get_viewset_for_model, get_request_without_query_params


class PermissionField(serializers.CharField):
    """Permission field."""

    def to_representation(self, obj):
        permission = getattr(obj, "permission", obj)
        return f"{permission.content_type.app_label}.{permission.codename}"


class ContentObjectField(serializers.CharField):
    """Content object field"""

    def to_representation(self, obj):
        content_object = getattr(obj, "content_object", None)
        if content_object:
            viewset = get_viewset_for_model(obj.content_object)

            request = get_request_without_query_params(self.context)

            serializer = viewset.serializer_class(obj.content_object, context={"request": request})
            return serializer.data.get("pulp_href")


class PermissionSerializer(serializers.Serializer):
    """Serializer for User/Group object permission."""

    pulp_href = serializers.SerializerMethodField(read_only=True)
    id = serializers.SerializerMethodField(read_only=True)
    permission = PermissionField(source="*", read_only=True)
    obj = ContentObjectField(help_text=_("Content object."), source="*", read_only=True)

    def get_id(self, obj) -> int:
        """Get model/object permission id."""
        return obj.id

    def get_pulp_href(self, obj) -> str:
        """Get model/object permission pulp_href."""
        group_pk = self.context.get("group_pk")

        if group_pk and isinstance(obj, Permission):
            return reverse("model_permissions-detail", args=[group_pk, obj.pk])

        if group_pk and isinstance(obj, GroupObjectPermission):
            return reverse("object_permissions-detail", args=[group_pk, obj.pk])

    def to_representation(self, obj):
        representation = super().to_representation(obj)

        if not self.context.get("group_pk"):
            representation.pop("pulp_href")

        return representation


class UserGroupSerializer(serializers.ModelSerializer):
    """Serializer for Groups that belong to an User."""

    name = serializers.CharField(help_text=_("Name."), max_length=150)
    pulp_href = IdentityField(view_name="groups-detail")

    class Meta:
        model = Group
        fields = ("name", "pulp_href")


class UserSerializer(serializers.ModelSerializer):
    """Serializer for User."""

    pulp_href = IdentityField(view_name="users-detail")
    id = serializers.IntegerField(read_only=True)
    username = serializers.CharField(
        help_text=_("Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only."),
        max_length=150,
    )
    first_name = serializers.CharField(help_text=_("First name"), max_length=150, allow_blank=True)
    last_name = serializers.CharField(help_text=_("Last name"), max_length=150, allow_blank=True)
    email = serializers.EmailField(help_text=_("Email address"), allow_blank=True)
    is_staff = serializers.BooleanField(
        help_text=_("Designates whether the user can log into this admin site."), default=False
    )
    is_active = serializers.BooleanField(
        help_text=_("Designates whether this user should be treated as active."), default=True
    )
    date_joined = serializers.DateTimeField(help_text=_("Date joined"), read_only=True)
    groups = UserGroupSerializer(read_only=True, many=True)

    class Meta:
        model = get_user_model()
        fields = (
            "pulp_href",
            "id",
            "username",
            "first_name",
            "last_name",
            "email",
            "is_staff",
            "is_active",
            "date_joined",
            "groups",
        )


class GroupUserSerializer(ValidateFieldsMixin, serializers.ModelSerializer):
    """Serializer for Users that belong to a Group."""

    username = serializers.CharField(
        help_text=_("Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only."),
        max_length=150,
    )
    pulp_href = IdentityField(view_name="users-detail")

    class Meta:
        model = get_user_model()
        fields = ("username", "pulp_href")


class GroupSerializer(ValidateFieldsMixin, serializers.ModelSerializer):
    """Serializer for Group."""

    pulp_href = IdentityField(view_name="groups-detail")
    id = serializers.IntegerField(read_only=True)
    name = serializers.CharField(help_text=_("Name"), max_length=150)

    class Meta:
        model = Group
        fields = ("name", "pulp_href", "id")



from unittest import mock

from django.test import TestCase

from pulpcore.download import BaseDownloader
from pulpcore.plugin.exceptions import UnsupportedDigestValidationError


class BaseDownloaderInitTestCase(TestCase):
    @mock.patch(
        "pulpcore.app.models.Artifact.DIGEST_FIELDS",
        new_callable=mock.PropertyMock,
        return_value=set(["sha512", "sha256"]),
    )
    def test_no_trusted_digest(self, mock_DIGEST_FIELDS):
        url = "http://example.com"
        with self.assertRaises(UnsupportedDigestValidationError):
            BaseDownloader(
                url,
                custom_file_object=None,
                expected_digests={"sha1": "912ec803b2ce49e4a541068d495ab570"},
                expected_size=None,
                semaphore=None,
            )

    @mock.patch(
        "pulpcore.app.models.Artifact.DIGEST_FIELDS",
        new_callable=mock.PropertyMock,
        return_value=set(["sha512", "sha256"]),
    )
    def test_no_expected_digests(self, mock_DIGEST_FIELDS):
        url = "http://example.com"
        downloader = BaseDownloader(
            url,
            custom_file_object=None,
            expected_digests=None,
            expected_size=None,
            semaphore=None,
        )
        assert downloader.expected_digests is None

    @mock.patch(
        "pulpcore.app.models.Artifact.DIGEST_FIELDS",
        new_callable=mock.PropertyMock,
        return_value=set(["sha512", "sha256"]),
    )
    def test_expected_digests(self, mock_DIGEST_FIELDS):
        url = "http://example.com"
        digests = {
            "sha1": "912ec803b2ce49e4a541068d495ab570",
            "sha256": "b361886f33f1b6089626dc8bd80961356f9a2911091ecb2ffa33730beb83bbdb",
        }
        downloader = BaseDownloader(
            url,
            custom_file_object=None,
            expected_digests=digests,
            expected_size=None,
            semaphore=None,
        )
        assert downloader.expected_digests == digests



import unittest
import uuid

from pulp_smash import api, cli, config
from pulp_smash.pulp3.bindings import monitor_task

from pulpcore.client.pulpcore import ApiClient, OrphansCleanupApi, TasksApi


class CorrelationIdTestCase(unittest.TestCase):
    """Test correlation id functionality."""

    @classmethod
    def setUpClass(cls):
        """Set up class variables."""
        cls.cid = str(uuid.uuid4())
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.client = ApiClient(
            configuration=cls.cfg.get_bindings_config(),
            header_name="Correlation-ID",
            header_value=cls.cid,
        )
        cls.cli_client = cli.Client(cls.cfg)

        cls.orphan_cleanup_api = OrphansCleanupApi(cls.client)
        cls.task_api = TasksApi(cls.client)

    def test_correlation_id(self):
        """Test that a correlation can be passed as a header and logged."""
        response, status, headers = self.orphan_cleanup_api.cleanup_with_http_info({})
        monitor_task(response.task)
        task = self.task_api.read(response.task)
        self.assertEqual(headers["Correlation-ID"], self.cid)
        self.assertEqual(task.logging_cid, self.cid)



"""A wrapper around `hashlib` providing only hashers named in settings.ALLOWED_CONTENT_CHECKSUMS"""

from gettext import gettext as _
import hashlib as the_real_hashlib

from django.conf import settings


def new(name, *args, **kwargs):
    """
    A wrapper around the real `hashlib.new()` providing only trusted hashers.

    The `ALLOWED_CONTENT_CHECKSUMS` setting identifies which hashers are allowed for use by Pulp.
    This function raises an exception if a hasher is requested which is not allowed, and otherwise,
    returns the standard hasher from `hashlib.new()`.

    Args:
        name: The name of the hasher to be instantiated.
        *args: args to be passed along to the real `hashlib.new()`.
        **kwargs: kwargs to be passed along to the real `hashlib.new()`

    Returns:
        An instantiated hasher, if it is allowed according to `ALLOWED_CONTENT_CHECKSUMS` setting.

    Raises:
        An exception if the name of the hasher is not in the `ALLOWED_CONTENT_CHECKSUMS` settings.

    """
    if name not in settings.ALLOWED_CONTENT_CHECKSUMS:
        raise Exception(
            _(
                "Hasher {} attempted to be used but is not in the `ALLOWED_CONTENT_CHECKSUMS` "
                "setting"
            ).format(name)
        )
    return the_real_hashlib.new(name, *args, **kwargs)



import os
import tempfile
from unittest import mock

from django.core.files.storage import default_storage as storage
from django.core.files.uploadedfile import SimpleUploadedFile

from django.conf import settings
from django.test import TestCase
from pulpcore.plugin.exceptions import (
    UnsupportedDigestValidationError,
    MissingDigestValidationError,
)

from pulpcore.plugin.models import (
    Artifact,
    Content,
    ContentArtifact,
    PulpTemporaryFile,
    Remote,
    RemoteArtifact,
)


class ContentCRUDTestCase(TestCase):

    artifact01_path = os.path.join(tempfile.gettempdir(), "artifact01-tmp")
    artifact02_path = os.path.join(tempfile.gettempdir(), "artifact02-tmp")

    def setUp(self):
        with open(self.artifact01_path, "w") as f:
            f.write("Temp Artifact File 01")
        with open(self.artifact02_path, "w") as f:
            f.write("Temp Artifact File 02")
        self.artifact01 = Artifact.init_and_validate(self.artifact01_path)
        self.artifact01.save()
        self.artifact02 = Artifact.init_and_validate(self.artifact02_path)
        self.artifact02.save()

    def test_create_and_read_content(self):
        content = Content.objects.create()
        content.save()
        artifact_file = storage.open(self.artifact01.file.name)
        content_artifact = ContentArtifact.objects.create(
            artifact=self.artifact01, content=content, relative_path=artifact_file.name
        )
        content_artifact.save()
        self.assertTrue(
            Content.objects.filter(pk=content.pk).exists()
            and ContentArtifact.objects.get(pk=content_artifact.pk).content.pk
            == Content.objects.get(pk=content.pk).pk
        )

    def test_remove_content(self):
        content = Content.objects.create()
        content.save()
        # Assumes creation is tested by test_create_and_read_content function
        Content.objects.filter(pk=content.pk).delete()
        self.assertFalse(Content.objects.filter(pk=content.pk).exists())


class PulpTemporaryFileTestCase(TestCase):
    def test_storage_location(self):
        if settings.DEFAULT_FILE_STORAGE != "pulpcore.app.models.storage.FileSystem":
            self.skipTest("Skipping test for nonlocal storage.")

        with tempfile.NamedTemporaryFile("ab") as tf:
            temp_file = PulpTemporaryFile(file=tf.name)
            temp_file.save()

        assert temp_file.file.name.startswith("tmp/files/")
        name = temp_file.file.file.name
        assert name.startswith("/var/lib/pulp/media/tmp/files"), name

    def test_read_temp_file(self):
        with tempfile.NamedTemporaryFile("ab") as tf:
            tf.write(b"temp file test")
            tf.flush()
            temp_file = PulpTemporaryFile(file=tf.name)
            temp_file.save()

        assert b"temp file test" in temp_file.file.read()


class ArtifactAlgorithmTestCase(TestCase):
    @mock.patch(
        "pulpcore.app.models.Artifact.FORBIDDEN_DIGESTS",
        new_callable=mock.PropertyMock,
        return_value=set(["md5"]),
    )
    @mock.patch(
        "pulpcore.app.models.Artifact.DIGEST_FIELDS",
        new_callable=mock.PropertyMock,
        return_value=set(["sha512", "sha384", "sha224", "sha1", "sha256"]),
    )
    def test_direct_set_forbidden(self, mock_FORBIDDEN_DIGESTS, mock_DIGEST_FIELDS):
        with self.assertRaises(UnsupportedDigestValidationError):
            a = Artifact(
                file=SimpleUploadedFile("test_filename", b"test content"),
                sha512="asdf",
                sha384="asdf",
                sha224="asdf",
                sha1="asdf",
                sha256="asdf",
                size=1024,
            )
            a.md5 = "asdf"
            a.save()

    @mock.patch(
        "pulpcore.app.models.Artifact.FORBIDDEN_DIGESTS",
        new_callable=mock.PropertyMock,
        return_value=set(["md5"]),
    )
    @mock.patch(
        "pulpcore.app.models.Artifact.DIGEST_FIELDS",
        new_callable=mock.PropertyMock,
        return_value=set(["sha512", "sha384", "sha224", "sha1", "sha256"]),
    )
    def test_forgot_something(self, mock_FORBIDDEN_DIGESTS, mock_DIGEST_FIELDS):
        with self.assertRaises(MissingDigestValidationError):
            a = Artifact(
                file=SimpleUploadedFile("test_filename", b"test content"),
                sha512="asdf",
                sha384="asdf",
                sha224="asdf",
                sha1="asdf",
                sha256="asdf",
                size=1024,
            )
            a.sha224 = None
            a.save()


@mock.patch(
    "pulpcore.app.models.Artifact.FORBIDDEN_DIGESTS",
    new_callable=mock.PropertyMock,
    return_value=set(["md5", "sha1"]),
)
@mock.patch(
    "pulpcore.app.models.Artifact.DIGEST_FIELDS",
    new_callable=mock.PropertyMock,
    return_value=set(["sha512", "sha384", "sha224", "sha256"]),
)
class RemoteArtifactAlgorithmTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.c = Content.objects.create()
        cls.ca = ContentArtifact.objects.create(artifact=None, content=cls.c, relative_path="ca")
        cls.remote = Remote.objects.create(url="http://example.org/")

    @classmethod
    def tearDownClass(cls):
        cls.ca.delete()
        cls.c.delete()
        cls.remote.delete()

    def test_remoteartifact_with_no_checksums(self, mock_FORBIDDEN_DIGESTS, mock_DIGEST_FIELDS):
        ra = RemoteArtifact(
            url="http://example.org/file",
            size=1024,
            md5=None,
            sha1=None,
            sha224=None,
            sha256="",
            sha384=None,
            sha512=None,
            content_artifact=self.ca,
            remote=self.remote,
        )
        ra.validate_checksums()

    def test_remoteartifact_with_allowed_checksums(
        self, mock_FORBIDDEN_DIGESTS, mock_DIGEST_FIELDS
    ):
        ra = RemoteArtifact(
            url="http://example.org/file",
            size=1024,
            md5="",
            sha1=None,
            sha224=None,
            sha256="sha256checksum",
            sha384=None,
            sha512=None,
            content_artifact=self.ca,
            remote=self.remote,
        )
        ra.validate_checksums()

    def test_remoteartifact_with_allowed_and_forbidden_checksums(
        self, mock_FORBIDDEN_DIGESTS, mock_DIGEST_FIELDS
    ):
        ra = RemoteArtifact(
            url="http://example.org/file",
            size=1024,
            md5="",
            sha1="sha1checksum",
            sha224=None,
            sha256="sha256checksum",
            sha384=None,
            sha512=None,
            content_artifact=self.ca,
            remote=self.remote,
        )
        ra.validate_checksums()

    def test_remoteartifact_with_forbidden_checksums(
        self, mock_FORBIDDEN_DIGESTS, mock_DIGEST_FIELDS
    ):
        with self.assertRaises(UnsupportedDigestValidationError):
            ra = RemoteArtifact(
                url="http://example.org/file",
                size=1024,
                md5="md5checksum",
                sha1=None,
                sha224=None,
                sha256="",
                sha384=None,
                sha512=None,
                content_artifact=self.ca,
                remote=self.remote,
            )
            ra.validate_checksums()



import gnupg
from pathlib import Path

from gettext import gettext as _

from django.core.management import BaseCommand, CommandError

from django.apps import apps
from django.db.utils import IntegrityError


class Command(BaseCommand):
    """
    Django management command for adding a signing service.

    This command is in tech-preview.
    """

    help = _("Adds a new AsciiArmoredDetachedSigningService. [tech-preview]")

    def add_arguments(self, parser):
        parser.add_argument(
            "name",
            help=_("Name, the signing_service should get in the database."),
        )
        parser.add_argument(
            "script",
            help=_("Shell script where the signing service is located."),
        )
        parser.add_argument(
            "key",
            help=_("Key id of the public key."),
        )
        parser.add_argument(
            "--class",
            default="core:AsciiArmoredDetachedSigningService",
            required=False,
            help=_("Signing service class prefixed by the app label separated by a colon."),
        )

    def handle(self, *args, **options):
        name = options["name"]
        script = options["script"]
        key_id = options["key"]

        if ":" not in options["class"]:
            raise CommandError(_("The signing service class was not provided in a proper format."))
        app_label, service_class = options["class"].split(":")

        try:
            SigningService = apps.get_model(app_label, service_class)
        except LookupError as e:
            raise CommandError(str(e))

        gpg = gnupg.GPG()
        key_list = gpg.list_keys(keys=[key_id])
        if not len(key_list) == 1:
            raise CommandError(_("There are {} keys matching the key id.").format(len(key_list)))
        fingerprint = key_list[0]["fingerprint"]
        public_key = gpg.export_keys(key_id)

        try:
            script_path = Path(script).resolve(strict=True)
        except FileNotFoundError as e:
            raise CommandError(str(e))

        try:
            SigningService.objects.create(
                name=name,
                public_key=public_key,
                pubkey_fingerprint=fingerprint,
                script=script_path,
            )
        except IntegrityError as e:
            raise CommandError(str(e))

        print(
            _("Successfully added signing service {name} for key {fingerprint}.").format(
                name=name, fingerprint=fingerprint
            )
        )



from drf_spectacular.utils import extend_schema
from drf_spectacular.utils import OpenApiParameter
from rest_framework import mixins
from rest_framework.decorators import action
from rest_framework.response import Response

from pulpcore.app import tasks
from pulpcore.app.models import Upload
from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.serializers import (
    AsyncOperationResponseSerializer,
    UploadChunkSerializer,
    UploadCommitSerializer,
    UploadSerializer,
    UploadDetailSerializer,
)
from pulpcore.app.viewsets.base import NamedModelViewSet
from pulpcore.tasking.tasks import dispatch


class UploadViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
):
    """View for chunked uploads."""

    endpoint_name = "uploads"
    queryset = Upload.objects.all()
    http_method_names = ["get", "post", "head", "put", "delete"]

    content_range_parameter = OpenApiParameter(
        name="Content-Range",
        location=OpenApiParameter.HEADER,
        required=True,
        type=str,
        description="The Content-Range header specifies the location of the file chunk "
        "within the file.",
    )

    def get_serializer_class(self):
        if self.action == "retrieve":
            return UploadDetailSerializer
        if self.action == "update":
            return UploadChunkSerializer
        if self.action == "commit":
            return UploadCommitSerializer
        return UploadSerializer

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.action == "update" and self.kwargs.get("pk"):
            context["upload"] = self.get_object()
        return context

    @extend_schema(
        summary="Upload a file chunk",
        request=UploadChunkSerializer,
        parameters=[content_range_parameter],
        responses={200: UploadSerializer},
    )
    def update(self, request, pk=None):
        """
        Upload a chunk for an upload.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        chunk = serializer.validated_data["file"]
        start = serializer.validated_data["start"]
        sha256 = serializer.validated_data.get("sha256")

        upload = self.get_object()
        upload.append(chunk, start, sha256)

        serializer = UploadSerializer(upload, context={"request": request})
        return Response(serializer.data)

    @extend_schema(
        summary="Finish an Upload",
        request=UploadCommitSerializer,
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"])
    def commit(self, request, pk):
        """
        Queues a Task that creates an Artifact, and the Upload gets deleted and cannot be re-used.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        sha256 = serializer.validated_data["sha256"]

        upload = self.get_object()
        task = dispatch(tasks.upload.commit, exclusive_resources=[upload], args=(upload.pk, sha256))
        return OperationPostponedResponse(task, request)



# Generated by Django 2.2.17 on 2021-01-05 21:18

from django.db import migrations, models
import django.db.models.deletion
import django_lifecycle.mixins
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('core', '0054_add_public_key'),
    ]

    operations = [
        migrations.CreateModel(
            name='Label',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('object_id', models.UUIDField()),
                ('key', models.CharField(max_length=200)),
                ('value', models.TextField(null=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'unique_together': {('content_type', 'object_id', 'key')},
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
    ]



"""Tests related to content promotion."""
import hashlib
import unittest
from urllib.parse import urljoin

from pulp_smash import api, config
from pulp_smash.pulp3.utils import gen_distribution, gen_remote, gen_repo, get_added_content, sync

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_CONTENT_NAME,
    FILE_DISTRIBUTION_PATH,
    FILE_FIXTURE_MANIFEST_URL,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
    PULP_CONTENT_BASE_URL,
)
from pulpcore.tests.functional.api.using_plugin.utils import create_file_publication
from pulpcore.tests.functional.api.using_plugin.utils import (  # noqa:F401
    set_up_module as setUpModule,
)


class ContentPromotionTestCase(unittest.TestCase):
    """Test content promotion."""

    def test_all(self):
        """Test content promotion for a distribution.

        This test targets the following issue:

        * `Pulp #4186 <https://pulp.plan.io/issues/4186>`_
        * `Pulp #8475 <https://pulp.plan.io/issues/8475>`_
        * `Pulp #8760 <https://pulp.plan.io/issues/8760>`_

        Do the following:

        1. Create a repository that has at least one repository version.
        2. Create a publication.
        3. Create 2 distributions - using the same publication. Those
           distributions will have different ``base_path``.
        4. Assert that distributions have the same publication.
        5. Create another distribution using same repository version.
        5. Assert that distributions are viewable from base url
        6. Assert that content in distributions are viewable
        7. Select a content unit. Download that content unit from Pulp using
           the three different distributions.
           Assert that content unit has the same checksum when fetched from
           different distributions.
        """
        cfg = config.get_config()
        client = api.Client(cfg, api.json_handler)

        repo = client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["pulp_href"])

        remote = client.post(FILE_REMOTE_PATH, gen_remote(FILE_FIXTURE_MANIFEST_URL))
        self.addCleanup(client.delete, remote["pulp_href"])

        sync(cfg, remote, repo)
        repo = client.get(repo["pulp_href"])

        publication = create_file_publication(cfg, repo)
        self.addCleanup(client.delete, publication["pulp_href"])

        distributions = []
        for _ in range(2):
            body = gen_distribution()
            body["publication"] = publication["pulp_href"]
            distribution = client.using_handler(api.task_handler).post(FILE_DISTRIBUTION_PATH, body)
            distributions.append(distribution)
            self.addCleanup(client.delete, distribution["pulp_href"])

        self.assertEqual(
            distributions[0]["publication"], distributions[1]["publication"], distributions
        )

        body = gen_distribution()
        body["repository"] = repo["pulp_href"]
        distribution = client.using_handler(api.task_handler).post(FILE_DISTRIBUTION_PATH, body)
        distributions.append(distribution)
        self.addCleanup(client.delete, distribution["pulp_href"])

        client.response_handler = api.safe_handler
        self.assertEqual(client.get(PULP_CONTENT_BASE_URL).status_code, 200)

        for distribution in distributions:
            self.assertEqual(client.get(distribution["base_url"]).status_code, 200)

        unit_urls = []
        unit_path = get_added_content(repo)[FILE_CONTENT_NAME][0]["relative_path"]
        for distribution in distributions:
            unit_url = distribution["base_url"]
            unit_urls.append(urljoin(unit_url, unit_path))

        self.assertEqual(
            hashlib.sha256(client.get(unit_urls[0]).content).hexdigest(),
            hashlib.sha256(client.get(unit_urls[1]).content).hexdigest(),
            unit_urls,
        )
        self.assertEqual(
            hashlib.sha256(client.get(unit_urls[0]).content).hexdigest(),
            hashlib.sha256(client.get(unit_urls[2]).content).hexdigest(),
            unit_urls,
        )



# Generated by Django 2.2.13 on 2020-06-30 14:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0034_groupprogressreport'),
    ]

    operations = [
        migrations.AddField(
            model_name='content',
            name='upstream_id',
            field=models.UUIDField(null=True),
        ),
    ]



from asgiref.sync import sync_to_async


# This is useful for querysets which don't have async support yet. Django querysets issue a db call
# when the iterator for them is requested, so we need that to be wrapped in `sync_to_async` also.
iter_async = sync_to_async(iter)


@sync_to_async
def next_async(it):
    try:
        return next(it)
    except StopIteration:
        raise StopAsyncIteration


def sync_to_async_iterable(sync_iterable):
    """
    Creates an async iterable.

    The returned iterator is able to be reused and iterated through multiple times.

    Args:
        sync_iterable: An iterable to be asynchronously iterated through.
    """

    class _Wrapper:
        def __aiter__(self):
            self.sync_iterator = None
            return self

        async def __anext__(self):
            if self.sync_iterator is None:
                self.sync_iterator = await iter_async(sync_iterable)
            return await next_async(self.sync_iterator)

    return _Wrapper()



import hashlib
import json
import os
import re
import subprocess
import tempfile
import tarfile
from gettext import gettext as _
from logging import getLogger

from django.core.files.storage import default_storage
from django.db.models import F

from pkg_resources import DistributionNotFound, get_distribution
from rest_framework.serializers import ValidationError
from tablib import Dataset

from pulpcore.app.apps import get_plugin_config
from pulpcore.app.models import (
    Artifact,
    Content,
    CreatedResource,
    GroupProgressReport,
    ProgressReport,
    PulpImport,
    PulpImporter,
    Repository,
    Task,
    TaskGroup,
)
from pulpcore.app.modelresource import (
    ArtifactResource,
    ContentArtifactResource,
)
from pulpcore.constants import TASK_STATES
from pulpcore.tasking.tasks import dispatch

log = getLogger(__name__)

ARTIFACT_FILE = "pulpcore.app.modelresource.ArtifactResource.json"
REPO_FILE = "pulpcore.app.modelresource.RepositoryResource.json"
CONTENT_FILE = "pulpcore.app.modelresource.ContentResource.json"
CA_FILE = "pulpcore.app.modelresource.ContentArtifactResource.json"
VERSIONS_FILE = "versions.json"
CONTENT_MAPPING_FILE = "content_mapping.json"


def _destination_repo(importer, source_repo_name):
    """Find the destination repository based on source repo's name."""
    if importer.repo_mapping and importer.repo_mapping.get(source_repo_name):
        dest_repo_name = importer.repo_mapping[source_repo_name]
    else:
        dest_repo_name = source_repo_name
    return Repository.objects.get(name=dest_repo_name)


def _import_file(fpath, resource_class, retry=False):
    try:
        log.info(_("Importing file {}.").format(fpath))
        with open(fpath, "r") as json_file:
            data = Dataset().load(json_file, format="json")
            resource = resource_class()
            log.info(_("...Importing resource {}.").format(resource.__class__.__name__))
            if retry:
                # django import-export can have a problem with concurrent-imports that are
                # importing the same 'thing' (e.g., a Package that exists in two different
                # repo-versions that are being imported at the same time). If we're asked to
                # retry, we will try an import that will simply record errors as they happen
                # (rather than failing with an exception) first. If errors happen, we'll do one
                # retry before we give up on this repo-version's import.
                a_result = resource.import_data(data, raise_errors=False)
                if a_result.has_errors():
                    log.info(
                        _("...{} import-errors encountered importing {}, retrying").format(
                            a_result.totals["error"], fpath
                        )
                    )
                    # Second attempt, we raise an exception on any problem.
                    # This will either succeed, or log a fatal error and fail.
                    try:
                        a_result = resource.import_data(data, raise_errors=True)
                    except Exception as e:  # noqa log on ANY exception and then re-raise
                        log.error(_("FATAL import-failure importing {}").format(fpath))
                        raise
            else:
                a_result = resource.import_data(data, raise_errors=True)
            return a_result
    except AttributeError:
        log.error(_("FAILURE importing file {}!").format(fpath))
        raise


def _check_versions(version_json):
    """Compare the export version_json to the installed components."""
    error_messages = []
    for component in version_json:
        try:
            version = get_distribution(component["component"]).version
        except DistributionNotFound:
            error_messages.append(
                _("Export uses {} which is not installed.").format(component["component"])
            )
        else:
            if version != component["version"]:
                error_messages.append(
                    _(
                        "Export version {export_ver} of {component} does not match "
                        "installed version {ver}."
                    ).format(
                        export_ver=component["version"],
                        component=component["component"],
                        ver=version,
                    )
                )

        if error_messages:
            raise ValidationError((" ".join(error_messages)))


def import_repository_version(importer_pk, destination_repo_pk, source_repo_name, tar_path):
    """
    Import a repository version from a Pulp export.

    Args:
        importer_pk (str): Importer we are working with
        destination_repo_pk (str): Primary key of Repository to import into.
        source_repo_name (str): Name of the Repository in the export.
        tar_path (str): A path to export tar.
    """
    dest_repo = Repository.objects.get(pk=destination_repo_pk)
    importer = PulpImporter.objects.get(pk=importer_pk)

    pb = ProgressReport(
        message=f"Importing content for {dest_repo.name}",
        code="import.repo.version.content",
        state=TASK_STATES.RUNNING,
    )
    pb.save()

    with tempfile.TemporaryDirectory() as temp_dir:
        # Extract the repo file for the repo info
        with tarfile.open(tar_path, "r:gz") as tar:
            tar.extract(REPO_FILE, path=temp_dir)

        with open(os.path.join(temp_dir, REPO_FILE), "r") as repo_data_file:
            data = json.load(repo_data_file)

        src_repo = next(repo for repo in data if repo["name"] == source_repo_name)

        if dest_repo.pulp_type != src_repo["pulp_type"]:
            raise ValidationError(
                _(
                    "Repository type mismatch: {src_repo} ({src_type}) vs {dest_repo} "
                    "({dest_type})."
                ).format(
                    src_repo=src_repo["name"],
                    src_type=src_repo["pulp_type"],
                    dest_repo=dest_repo.name,
                    dest_type=dest_repo.pulp_type,
                )
            )

        rv_name = ""
        # Extract the repo version files
        with tarfile.open(tar_path, "r:gz") as tar:
            for mem in tar.getmembers():
                match = re.search(rf"(^repository-{source_repo_name}_[0-9]+)/.+", mem.name)
                if match:
                    rv_name = match.group(1)
                    tar.extract(mem, path=temp_dir)

        if not rv_name:
            raise ValidationError(_("No RepositoryVersion found for {}").format(rv_name))

        rv_path = os.path.join(temp_dir, rv_name)
        # Content
        plugin_name = src_repo["pulp_type"].split(".")[0]
        cfg = get_plugin_config(plugin_name)

        resulting_content_ids = []
        for res_class in cfg.exportable_classes:
            filename = f"{res_class.__module__}.{res_class.__name__}.json"
            a_result = _import_file(os.path.join(rv_path, filename), res_class, retry=True)
            resulting_content_ids.extend(
                row.object_id for row in a_result.rows if row.import_type in ("new", "update")
            )

        # Once all content exists, create the ContentArtifact links
        ca_path = os.path.join(rv_path, CA_FILE)
        _import_file(ca_path, ContentArtifactResource, retry=True)

        # see if we have a content mapping
        mapping_path = f"{rv_name}/{CONTENT_MAPPING_FILE}"
        mapping = {}
        with tarfile.open(tar_path, "r:gz") as tar:
            if mapping_path in tar.getnames():
                tar.extract(mapping_path, path=temp_dir)
                with open(os.path.join(temp_dir, mapping_path), "r") as mapping_file:
                    mapping = json.load(mapping_file)

        if mapping:
            # use the content mapping to map content to repos
            for repo_name, content_ids in mapping.items():
                repo = _destination_repo(importer, repo_name)
                content = Content.objects.filter(upstream_id__in=content_ids)
                with repo.new_version() as new_version:
                    new_version.set_content(content)
        else:
            # just map all the content to our destination repo
            content = Content.objects.filter(pk__in=resulting_content_ids)
            with dest_repo.new_version() as new_version:
                new_version.set_content(content)

        content_count = content.count()
        pb.total = content_count
        pb.done = content_count
        pb.state = TASK_STATES.COMPLETED
        pb.save()

    gpr = TaskGroup.current().group_progress_reports.filter(code="import.repo.versions")
    gpr.update(done=F("done") + 1)


def pulp_import(importer_pk, path, toc):
    """
    Import a Pulp export into Pulp.

    Args:
        importer_pk (str): Primary key of PulpImporter to do the import
        path (str): Path to the export to be imported
    """

    def _compute_hash(filename):
        sha256_hash = hashlib.sha256()
        with open(filename, "rb") as f:
            # Read and update hash string value in blocks of 4K
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)
            return sha256_hash.hexdigest()

    def validate_toc(toc_filename):
        """
        Check validity of table-of-contents file.

        table-of-contents must:
          * exist
          * be valid JSON
          * point to chunked-export-files that exist 'next to' the 'toc' file
          * point to chunks whose checksums match the checksums stored in the 'toc' file

        Args:
            toc_filename (str): The user-provided toc-file-path to be validated.

        Raises:
            ValidationError: If toc is not a valid JSON table-of-contents file,
            or when toc points to chunked-export-files that can't be found in the same
            directory as the toc-file, or the checksums of the chunks do not match the
            checksums stored in toc.
        """
        with open(toc_filename) as json_file:
            # Valid JSON?
            the_toc = json.load(json_file)
            if not the_toc.get("files", None) or not the_toc.get("meta", None):
                raise ValidationError(_("Missing 'files' or 'meta' keys in table-of-contents!"))

            base_dir = os.path.dirname(toc_filename)
            # Points at chunks that exist?
            missing_files = []
            for f in sorted(the_toc["files"].keys()):
                if not os.path.isfile(os.path.join(base_dir, f)):
                    missing_files.append(f)
            if missing_files:
                raise ValidationError(
                    _(
                        "Missing import-chunks named in table-of-contents: {}.".format(
                            str(missing_files)
                        )
                    )
                )

            errs = []
            # validate the sha256 of the toc-entries
            # gather errors for reporting at the end
            chunks = sorted(the_toc["files"].keys())
            data = dict(message="Validating Chunks", code="validate.chunks", total=len(chunks))
            with ProgressReport(**data) as pb:
                for chunk in pb.iter(chunks):
                    a_hash = _compute_hash(os.path.join(base_dir, chunk))
                    if not a_hash == the_toc["files"][chunk]:
                        err_str = "File {} expected checksum : {}, computed checksum : {}".format(
                            chunk, the_toc["files"][chunk], a_hash
                        )
                        errs.append(err_str)

            # if there are any errors, report and fail
            if errs:
                raise ValidationError(_("Import chunk hash mismatch: {}).").format(str(errs)))

        return the_toc

    def validate_and_assemble(toc_filename):
        """Validate checksums of, and reassemble, chunks in table-of-contents file."""
        the_toc = validate_toc(toc_filename)
        toc_dir = os.path.dirname(toc_filename)
        result_file = os.path.join(toc_dir, the_toc["meta"]["file"])

        # if we have only one entry in "files", it must be the full .tar.gz - return it
        if len(the_toc["files"]) == 1:
            return os.path.join(toc_dir, list(the_toc["files"].keys())[0])

        # We have multiple chunks.
        # reassemble into one file 'next to' the toc and return the resulting full-path
        chunk_size = int(the_toc["meta"]["chunk_size"])
        offset = 0
        block_size = 1024
        blocks_per_chunk = int(chunk_size / block_size)

        # sorting-by-filename is REALLY IMPORTANT here
        # keys are of the form <base-export-name>.00..<base-export-name>.NN,
        # and must be reassembled IN ORDER
        the_chunk_files = sorted(the_toc["files"].keys())

        data = dict(
            message="Recombining Chunks", code="recombine.chunks", total=len(the_chunk_files)
        )
        with ProgressReport(**data) as pb:
            for chunk in pb.iter(the_chunk_files):
                # For each chunk, add it to the reconstituted tar.gz, picking up where the previous
                # chunk left off
                subprocess.run(
                    [
                        "dd",
                        "if={}".format(os.path.join(toc_dir, chunk)),
                        "of={}".format(result_file),
                        "bs={}".format(str(block_size)),
                        "seek={}".format(str(offset)),
                    ],
                )
                offset += blocks_per_chunk
                # To keep from taking up All The Disk, we delete each chunk after it has been added
                # to the recombined file.
                try:
                    subprocess.run(["rm", "-f", os.path.join(toc_dir, chunk)])
                except OSError:
                    log.warning(
                        _("Failed to remove chunk {} after recombining. Continuing.").format(
                            os.path.join(toc_dir, chunk)
                        ),
                        exc_info=True,
                    )

        combined_hash = _compute_hash(result_file)
        if combined_hash != the_toc["meta"]["global_hash"]:
            raise ValidationError(
                _("Mismatch between combined .tar.gz checksum [{}] and originating [{}]).").format(
                    combined_hash, the_toc["meta"]["global_hash"]
                )
            )
        # if we get this far, then: the chunk-files all existed, they all pass checksum validation,
        # and there exists a combined .tar.gz, which *also* passes checksum-validation.
        # Let the rest of the import process do its thing on the new combined-file.
        return result_file

    if toc:
        log.info(_("Validating TOC {}.").format(toc))
        path = validate_and_assemble(toc)

    log.info(_("Importing {}.").format(path))
    current_task = Task.current()
    task_group = TaskGroup.current()
    importer = PulpImporter.objects.get(pk=importer_pk)
    the_import = PulpImport.objects.create(
        importer=importer, task=current_task, params={"path": path}
    )
    CreatedResource.objects.create(content_object=the_import)

    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        with tarfile.open(path, "r:gz") as tar:
            tar.extractall(path=temp_dir)

        # Check version info
        with open(os.path.join(temp_dir, VERSIONS_FILE)) as version_file:
            version_json = json.load(version_file)
            _check_versions(version_json)

        # Artifacts
        ar_result = _import_file(os.path.join(temp_dir, ARTIFACT_FILE), ArtifactResource)
        data = dict(
            message="Importing Artifacts", code="import.artifacts", total=len(ar_result.rows)
        )
        with ProgressReport(**data) as pb:
            for row in pb.iter(ar_result.rows):
                artifact = Artifact.objects.get(pk=row.object_id)
                base_path = os.path.join("artifact", artifact.sha256[0:2], artifact.sha256[2:])
                src = os.path.join(temp_dir, base_path)

                if not default_storage.exists(base_path):
                    with open(src, "rb") as f:
                        default_storage.save(base_path, f)

        with open(os.path.join(temp_dir, REPO_FILE), "r") as repo_data_file:
            data = json.load(repo_data_file)
            gpr = GroupProgressReport(
                message="Importing repository versions",
                code="import.repo.versions",
                total=len(data),
                done=0,
                task_group=task_group,
            )
            gpr.save()

            for src_repo in data:
                try:
                    dest_repo = _destination_repo(importer, src_repo["name"])
                except Repository.DoesNotExist:
                    log.warning(
                        _("Could not find destination repo for {}. Skipping.").format(
                            src_repo["name"]
                        )
                    )
                    continue

                dispatch(
                    import_repository_version,
                    exclusive_resources=[dest_repo],
                    args=[importer.pk, dest_repo.pk, src_repo["name"], path],
                    task_group=task_group,
                )

    task_group.finish()



from gettext import gettext as _

from django.db import transaction
from rest_framework import serializers
from rest_framework.validators import UniqueValidator

from pulpcore.app import models
from pulpcore.app.serializers import base, fields


class BaseContentSerializer(base.ModelSerializer):
    pulp_href = base.DetailIdentityField(view_name_pattern=r"contents(-.*/.*)-detail")

    class Meta:
        model = models.Content
        fields = base.ModelSerializer.Meta.fields


class NoArtifactContentSerializer(BaseContentSerializer):
    class Meta:
        model = models.Content
        fields = BaseContentSerializer.Meta.fields


class SingleArtifactContentSerializer(BaseContentSerializer):
    artifact = fields.SingleContentArtifactField(
        help_text=_("Artifact file representing the physical content"),
    )

    relative_path = serializers.CharField(
        help_text=_("Path where the artifact is located relative to distributions base_path"),
        validators=[fields.relative_path_validator],
        write_only=True,
    )

    def __init__(self, *args, **kwargs):
        """
        Initializer for SingleArtifactContentSerializer
        """
        super().__init__(*args, **kwargs)

        # If the content model has its own database field 'relative_path',
        # we should not mark the field write_only
        if hasattr(self.Meta.model, "relative_path") and "relative_path" in self.fields:
            self.fields["relative_path"].write_only = False

    @transaction.atomic
    def create(self, validated_data):
        """
        Create the content and associate it with its Artifact.

        Args:
            validated_data (dict): Data to save to the database
        """
        artifact = validated_data.pop("artifact")
        if "relative_path" not in self.fields or self.fields["relative_path"].write_only:
            relative_path = validated_data.pop("relative_path")
        else:
            relative_path = validated_data.get("relative_path")
        content = self.Meta.model.objects.create(**validated_data)
        models.ContentArtifact.objects.create(
            artifact=artifact, content=content, relative_path=relative_path
        )
        return content

    class Meta:
        model = models.Content
        fields = BaseContentSerializer.Meta.fields + ("artifact", "relative_path")


class MultipleArtifactContentSerializer(BaseContentSerializer):
    artifacts = fields.ContentArtifactsField(
        help_text=_(
            "A dict mapping relative paths inside the Content to the corresponding"
            "Artifact URLs. E.g.: {'relative/path': "
            "'/artifacts/1/'"
        ),
    )

    @transaction.atomic
    def create(self, validated_data):
        """
        Create the content and associate it with all its Artifacts.

        Args:
            validated_data (dict): Data to save to the database
        """
        artifacts = validated_data.pop("artifacts")
        content = self.Meta.model.objects.create(**validated_data)
        for relative_path, artifact in artifacts.items():
            models.ContentArtifact.objects.create(
                artifact=artifact, content=content, relative_path=relative_path
            )
        return content

    class Meta:
        model = models.Content
        fields = BaseContentSerializer.Meta.fields + ("artifacts",)


class ContentChecksumSerializer(serializers.Serializer):
    """
    Provide a serializer with artifact checksum fields for single artifact content.

    If you use this serializer, it's recommended that you prefetch artifacts:

        Content.objects.prefetch_related("_artifacts").all()
    """

    md5 = fields.ContentArtifactChecksumField(
        help_text=_("The MD5 checksum if available."),
        checksum="md5",
    )

    sha1 = fields.ContentArtifactChecksumField(
        help_text=_("The SHA-1 checksum if available."),
        checksum="sha1",
    )

    sha224 = fields.ContentArtifactChecksumField(
        help_text=_("The SHA-224 checksum if available."),
        checksum="sha224",
    )

    sha256 = fields.ContentArtifactChecksumField(
        help_text=_("The SHA-256 checksum if available."),
        checksum="sha256",
    )

    sha384 = fields.ContentArtifactChecksumField(
        help_text=_("The SHA-384 checksum if available."),
        checksum="sha384",
    )

    sha512 = fields.ContentArtifactChecksumField(
        help_text=_("The SHA-512 checksum if available."),
        checksum="sha512",
    )

    class Meta:
        model = models.Content
        fields = base.ModelSerializer.Meta.fields + (
            "md5",
            "sha1",
            "sha224",
            "sha256",
            "sha384",
            "sha512",
        )


class ArtifactSerializer(base.ModelSerializer):
    pulp_href = base.IdentityField(view_name="artifacts-detail")

    file = serializers.FileField(help_text=_("The stored file."), allow_empty_file=True)

    size = serializers.IntegerField(help_text=_("The size of the file in bytes."), required=False)

    md5 = serializers.CharField(
        help_text=_("The MD5 checksum of the file if available."), required=False, allow_null=True
    )

    sha1 = serializers.CharField(
        help_text=_("The SHA-1 checksum of the file if available."),
        required=False,
        allow_null=True,
    )

    sha224 = serializers.CharField(
        help_text=_("The SHA-224 checksum of the file if available."),
        required=False,
        allow_null=True,
    )

    sha256 = serializers.CharField(
        help_text=_("The SHA-256 checksum of the file if available."),
        required=False,
        allow_null=True,
    )

    sha384 = serializers.CharField(
        help_text=_("The SHA-384 checksum of the file if available."),
        required=False,
        allow_null=True,
    )

    sha512 = serializers.CharField(
        help_text=_("The SHA-512 checksum of the file if available."),
        required=False,
        allow_null=True,
    )

    def validate(self, data):
        """
        Validate file by size and by all checksums provided.

        Args:
            data (:class:`django.http.QueryDict`): QueryDict mapping Artifact model fields to their
                values

        Raises:
            :class:`rest_framework.exceptions.ValidationError`: When the expected file size or any
                of the checksums don't match their actual values.
        """
        super().validate(data)
        if "size" in data:
            if data["file"].size != int(data["size"]):
                raise serializers.ValidationError(_("The size did not match actual size of file."))
        else:
            data["size"] = data["file"].size

        bad_algs = []
        for algorithm in models.Artifact.FORBIDDEN_DIGESTS:
            if algorithm in data:
                bad_algs.append(algorithm)
        if bad_algs:
            raise serializers.ValidationError(
                _("Checksum algorithms {} forbidden for this Pulp instance.").format(bad_algs)
            )

        for algorithm in reversed(models.Artifact.DIGEST_FIELDS):
            digest = data["file"].hashers[algorithm].hexdigest()

            if algorithm in data and digest != data[algorithm]:
                raise serializers.ValidationError(_("The %s checksum did not match.") % algorithm)
            else:
                data[algorithm] = digest

            if algorithm in models.Artifact.RELIABLE_DIGEST_FIELDS:
                validator = UniqueValidator(
                    models.Artifact.objects.all(),
                    message=_("Artifact with {0} checksum of '{1}' already exists.").format(
                        algorithm, digest
                    ),
                )
                validator.instance = None
                validator(digest, self.fields[algorithm])

        return data

    class Meta:
        model = models.Artifact
        fields = base.ModelSerializer.Meta.fields + (
            "file",
            "size",
            "md5",
            "sha1",
            "sha224",
            "sha256",
            "sha384",
            "sha512",
        )


class SigningServiceSerializer(base.ModelSerializer):
    """
    A serializer for the model declaring a signing service.
    """

    pulp_href = base.IdentityField(view_name="signing-services-detail")
    name = serializers.CharField(help_text=_("A unique name used to recognize a script."))
    public_key = serializers.CharField(
        help_text=_("The value of a public key used for the repository verification.")
    )
    pubkey_fingerprint = serializers.CharField(help_text=_("The fingerprint of the public key."))
    script = serializers.CharField(
        help_text=_("An absolute path to a script which is going to be used for the signing.")
    )

    class Meta:
        model = models.SigningService
        fields = BaseContentSerializer.Meta.fields + (
            "name",
            "public_key",
            "pubkey_fingerprint",
            "script",
        )



"""
Repository related Django models.
"""
from django.db import models

from .base import MasterModel, BaseModel


class AlternateContentSource(MasterModel):
    """
    Alternate sources of content.

    Fields:

        name (models.TextField): The alternate content source name.
        last_refreshed (models.DateTimeField): Last refreshed date.
        url (models.TextField): URL of Alternate Content Source.

    Relations:

        remote (models.ForeignKeyField): Associated remote
    """

    TYPE = "acs"
    REMOTE_TYPES = []

    name = models.TextField(db_index=True, unique=True)
    last_refreshed = models.DateTimeField(null=True)
    remote = models.ForeignKey("Remote", null=True, on_delete=models.PROTECT)

    class Meta:
        verbose_name_plural = "acs"


class AlternateContentSourcePath(BaseModel):
    """
    Alternate sources of content.

    Fields:

        path (models.TextField): The alternate content source name.

    Relations:

        alternate_content_source (models.ForeignKeyField): Associated AlternateContentSource
        repository (models.ForeignKeyField): Associated repository
    """

    alternate_content_source = models.ForeignKey(
        "AlternateContentSource", null=True, on_delete=models.CASCADE, related_name="paths"
    )
    path = models.TextField(default=None)
    repository = models.ForeignKey("Repository", null=True, on_delete=models.CASCADE)

    class Meta:
        unique_together = ("alternate_content_source", "path")



"""Tests related to the AccessPolicy."""

import unittest

from pulp_smash import config

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    AccessPoliciesApi,
)


class AccessPolicyTestCase(unittest.TestCase):
    """
    Test cases for AccessPolicy.
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.core_client = CoreApiClient(configuration=cls.cfg.get_bindings_config())
        cls.access_policy_api = AccessPoliciesApi(cls.core_client)

    def test_access_policy_cannot_be_created(self):
        """Test that only plugin writers can ship a new AccessPolicy."""
        self.assertFalse(hasattr(self.access_policy_api, "create"))

    def test_access_policy_default_policies(self):
        """Test that the default policies from pulpcore are installed."""
        groups_response = self.access_policy_api.list(viewset_name="groups")
        self.assertEqual(groups_response.count, 1)

        groups_users_response = self.access_policy_api.list(viewset_name="groups/users")
        self.assertEqual(groups_users_response.count, 1)

        tasks_response = self.access_policy_api.list(viewset_name="tasks")
        self.assertEqual(tasks_response.count, 1)

    def test_statements_attr_can_be_modified(self):
        """Test that `AccessPolicy.statements` can be modified"""
        tasks_response = self.access_policy_api.list(viewset_name="tasks")
        tasks_href = tasks_response.results[0].pulp_href
        task_access_policy = self.access_policy_api.read(tasks_href)

        original_statements = task_access_policy.statements
        self.assertFalse(task_access_policy.customized)
        self.assertNotEqual(original_statements, [])

        self.access_policy_api.partial_update(tasks_href, {"statements": []})
        task_access_policy = self.access_policy_api.read(tasks_href)
        self.assertTrue(task_access_policy.customized)
        self.assertEqual(task_access_policy.statements, [])

        self.access_policy_api.partial_update(tasks_href, {"statements": original_statements})
        task_access_policy = self.access_policy_api.read(tasks_href)
        self.assertTrue(task_access_policy.customized)
        self.assertEqual(task_access_policy.statements, original_statements)

    def test_permissions_assignment_attr_can_be_modified(self):
        """Test that `AccessPolicy.permissions_assignment` can be modified"""
        groups_response = self.access_policy_api.list(viewset_name="groups")
        groups_href = groups_response.results[0].pulp_href
        groups_access_policy = self.access_policy_api.read(groups_href)

        original_permissions_assignment = groups_access_policy.permissions_assignment
        self.assertFalse(groups_access_policy.customized)
        self.assertNotEqual(original_permissions_assignment, [])

        self.access_policy_api.partial_update(groups_href, {"permissions_assignment": []})
        groups_access_policy = self.access_policy_api.read(groups_href)
        self.assertTrue(groups_access_policy.customized)
        self.assertEqual(groups_access_policy.permissions_assignment, [])

        self.access_policy_api.partial_update(
            groups_href, {"permissions_assignment": original_permissions_assignment}
        )
        groups_access_policy = self.access_policy_api.read(groups_href)
        self.assertTrue(groups_access_policy.customized)
        self.assertEqual(
            groups_access_policy.permissions_assignment, original_permissions_assignment
        )

    def test_customized_is_read_only(self):
        """Test that the `AccessPolicy.customized` attribute is read only"""
        tasks_response = self.access_policy_api.list(viewset_name="tasks")
        tasks_href = tasks_response.results[0].pulp_href
        task_access_policy = self.access_policy_api.read(tasks_href)

        response = self.access_policy_api.partial_update(
            tasks_href, {"customized": not task_access_policy.customized}
        )
        self.assertEqual(response.customized, task_access_policy.customized)

    def test_viewset_name_is_read_only(self):
        """Test that the `AccessPolicy.viewset_name` attribute is read only"""
        tasks_response = self.access_policy_api.list(viewset_name="tasks")
        tasks_href = tasks_response.results[0].pulp_href
        task_access_policy = self.access_policy_api.read(tasks_href)

        response = self.access_policy_api.partial_update(
            tasks_href, {"viewset_name": "not-a-real-name"}
        )
        self.assertEqual(response.viewset_name, task_access_policy.viewset_name)
